import pytest

import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave import box
from weave.legacy.weave.arrow.arrow import ArrowWeaveListType
from weave.legacy.weave.language_features.tagging import tag_store, tagged_value_type
from weave.legacy.tests.util.list_arrow_test_helpers import ArrowNode

pick_options = [
    # Basic Pick
    ({"a": 1}, "a", weave.types.Int(), 1),
    # Missing Key
    ({"a": 1}, "b", weave.types.NoneType(), None),
    # Partially Available Mixed Key
    (
        {"a": [{"a": 1}, {"b": 2}, {"b": 3}, {"b": "hi"}]},
        "a.*.b",
        weave.types.List(
            weave.types.union(
                weave.types.Int(), weave.types.String(), weave.types.NoneType()
            )
        ),
        [None, 2, 3, "hi"],
    ),
    # Complicated Case
    (
        {
            "extra": "data",
            "outer": [
                {
                    "extra": "data",
                    "b": [
                        [{"extra": "data", "c": {"extra": "data", ".d": [1, 2, 3]}}],
                        [{"extra": "data", "c": {"extra": "data", ".d": [11, 12, 13]}}],
                    ],
                },
                {
                    "extra": "data",
                    "b": [
                        [{"extra": "data", "c": {"extra": "data", ".d": [21, 22, 23]}}],
                        [{"extra": "data", "c": {"extra": "data", ".d": [31, 32, 33]}}],
                    ],
                },
            ],
        },
        "outer.*.b.*.*.c.\\.d.*",
        weave.types.List(
            weave.types.List(weave.types.List(weave.types.List(weave.types.Int())))
        ),
        [[[[1, 2, 3]], [[11, 12, 13]]], [[[21, 22, 23]], [[31, 32, 33]]]],
    ),
]


@pytest.mark.parametrize("val, pick_key, exp_type, exp_val", pick_options)
def test_simple_typeddict_pick(val, pick_key, exp_type, exp_val):
    # Flat:
    node = weave_internal.const(val).pick(pick_key)
    assert node.type == exp_type
    assert weave.use(node) == exp_val


@pytest.mark.parametrize("val, pick_key, exp_type, exp_val", pick_options)
def test_list_typeddict_pick(val, pick_key, exp_type, exp_val):
    # Inside a list:
    node = weave_internal.const([val, val]).pick(pick_key)
    assert node.type == weave.types.List(exp_type)
    assert weave.use(node) == [exp_val, exp_val]


awl_pick_options = pick_options.copy()
awl_pick_options[
    2
] = (  # We have to modify this last one because arrow does not support sparse unions
    {"a": [{"a": 1}, {"b": 2}, {"b": 3}]},
    "a.*.b",
    weave.types.List(weave.types.union(weave.types.Int(), weave.types.NoneType())),
    [None, 2, 3],
)  # type: ignore


@pytest.mark.parametrize("val, pick_key, exp_type, exp_val", awl_pick_options)
def test_awl_typeddict_pick(val, pick_key, exp_type, exp_val):
    # Inside a AWL:
    node = ArrowNode.make_node([val, val]).pick(pick_key)
    assert node.type == ArrowWeaveListType(exp_type)
    assert ArrowNode.use_node(node) == [exp_val, exp_val]


def test_pick_on_crazy_nested_tagged_obj():
    def t(val, tag):
        return tag_store.add_tags(box.box(val), tag)

    def tt(tag, inner):
        return tagged_value_type.TaggedValueType(weave.types.TypedDict(tag), inner)

    val = {
        "extra": "data",
        "outer": t(
            [
                t(
                    {
                        "extra": "data",
                        "b": t(
                            [
                                t(
                                    [
                                        t(
                                            {
                                                "extra": "data",
                                                "c": t(
                                                    {
                                                        "extra": "data",
                                                        ".d": t(
                                                            [
                                                                t(
                                                                    0,
                                                                    {
                                                                        "o_*_b_*_*_c_d_*": 0
                                                                    },
                                                                ),
                                                                t(
                                                                    1,
                                                                    {
                                                                        "o_*_b_*_*_c_d_*": 1
                                                                    },
                                                                ),
                                                                t(
                                                                    2,
                                                                    {
                                                                        "o_*_b_*_*_c_d_*": 2
                                                                    },
                                                                ),
                                                            ],
                                                            {"o_*_b_*_*_c_d": 3},
                                                        ),
                                                    },
                                                    {"o_*_b_*_*_c": 4},
                                                ),
                                            },
                                            {"o_*_b_*_*": 5},
                                        )
                                    ],
                                    {"o_*_b_*": 6},
                                ),
                            ],
                            {"o_*_b": 7},
                        ),
                    },
                    {"o_*": 8},
                ),
            ],
            {"o": 9},
        ),
    }
    pick_key = "outer.*.b.*.*.c.\\.d.*"
    it = weave.types.Int()
    lt = weave.types.List
    exp_type = tt(
        {"o": it},
        lt(
            tt(
                {"o_*": it, "o_*_b": it},
                lt(
                    tt(
                        {"o_*_b_*": it},
                        lt(
                            tt(
                                {
                                    "o_*_b_*_*": it,
                                    "o_*_b_*_*_c": it,
                                    "o_*_b_*_*_c_d": it,
                                },
                                lt(tt({"o_*_b_*_*_c_d_*": it}, it)),
                            )
                        ),
                    )
                ),
            )
        ),
    )
    weave.types.List(
        weave.types.List(weave.types.List(weave.types.List(weave.types.Int())))
    )
    exp_val = [[[[0, 1, 2]]]]

    # Flat:
    node = weave_internal.const(val).pick(pick_key)
    assert node.type == exp_type
    assert weave.use(node) == exp_val

    # Inside a list:
    node = weave_internal.const([val, val]).pick(pick_key)
    assert node.type == weave.types.List(exp_type)
    assert weave.use(node) == [exp_val, exp_val]

    # Inside a AWL:
    node = ArrowNode.make_node([val, val]).pick(pick_key)
    assert node.type == ArrowWeaveListType(exp_type)
    assert ArrowNode.use_node(node) == [exp_val, exp_val]



import pytest

from weave.integrations.integration_utilities import truncate_op_name

MAX_RUN_NAME_LENGTH = 128
NON_HASH_LIMIT = 5


def _make_string_of_length(n: int) -> str:
    return "a" * n


def test_truncate_op_name_less_than_limit() -> None:
    name = _make_string_of_length(MAX_RUN_NAME_LENGTH - 1)
    trunc = truncate_op_name(name)
    assert trunc == name


def test_truncate_op_name_at_limit() -> None:
    name = _make_string_of_length(MAX_RUN_NAME_LENGTH)
    trunc = truncate_op_name(name)
    assert trunc == name


def _truncated_str(tail_len: int, total_len: int) -> tuple:
    name = (
        _make_string_of_length(total_len - tail_len - 1)
        + "."
        + _make_string_of_length(tail_len)
    )
    return name, truncate_op_name(name)


def test_truncate_op_name_too_short_for_hash() -> None:
    # Remove 1 character for a range of tail lengths:
    chars_to_remove = 1
    for tail_len in range(NON_HASH_LIMIT + 1):
        if tail_len <= chars_to_remove:
            with pytest.raises(ValueError):
                name, trunc = _truncated_str(
                    tail_len, MAX_RUN_NAME_LENGTH + chars_to_remove
                )
        else:
            name, trunc = _truncated_str(
                tail_len, MAX_RUN_NAME_LENGTH + chars_to_remove
            )
            assert trunc == name[:MAX_RUN_NAME_LENGTH]

    name, trunc = _truncated_str(NON_HASH_LIMIT + 1, MAX_RUN_NAME_LENGTH + 1)
    assert (
        trunc
        == "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa.a_0_a"
    )

    # Remove a range of character for a fixed tail length:
    tail_len = NON_HASH_LIMIT
    for chars_to_remove in range(0, tail_len + 1):
        if tail_len <= chars_to_remove:
            with pytest.raises(ValueError):
                name, trunc = _truncated_str(
                    tail_len, MAX_RUN_NAME_LENGTH + chars_to_remove
                )
        else:
            name, trunc = _truncated_str(
                tail_len, MAX_RUN_NAME_LENGTH + chars_to_remove
            )
            assert trunc == name[:MAX_RUN_NAME_LENGTH]


def test_truncate_op_name_with_digest() -> None:
    name = _make_string_of_length(MAX_RUN_NAME_LENGTH + 1)
    trunc = truncate_op_name(name)
    assert (
        trunc
        == "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa_b325_aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    )

    name = _make_string_of_length(MAX_RUN_NAME_LENGTH + 10)
    trunc = truncate_op_name(name)
    assert (
        trunc
        == "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa_55b6_aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    )



# Tests interactions between language features.

import typing

import weave
from weave.legacy.weave import box, context_state
from weave.legacy.weave.language_features.tagging import (
    make_tag_getter_op,
    tag_store,
    tagged_value_type,
)

_loading_builtins_token = context_state.set_loading_built_ins()


@weave.op(
    output_type=lambda input_types: tagged_value_type.TaggedValueType(
        weave.types.TypedDict({"a_tag": weave.types.Int()}),
        input_types["x"],
    ),
    hidden=True,
)
def _test_op_tag_input(x: typing.Any, tag_val: int):
    x = box.box(x)
    tag_store.add_tags(x, {"a_tag": tag_val})
    return x


get_a_tag = make_tag_getter_op.make_tag_getter_op(
    "a_tag", weave.types.Int(), op_name="tag-a_tag"
)


@weave.op(hidden=True)
def _test_op_refining_refine(x: typing.Any) -> weave.types.Type:
    return weave.type_of(x)


@weave.op(refine_output_type=_test_op_refining_refine, hidden=True)
def _test_op_refining(x: typing.Any) -> typing.Any:
    return x


context_state.clear_loading_built_ins(_loading_builtins_token)


def test_none():
    node = weave.save(None, "null")
    x = node + 1
    assert x.type == weave.types.NoneType()
    assert weave.use(x) == None


def test_empty_list():
    node = weave.save([], "null")
    x = node + 1
    assert x.type == weave.types.List(weave.types.Number())
    assert weave.use(x) == []


def test_list_none():
    node = weave.save([None, None], "null")
    x = node + 1
    assert x.type == weave.types.List(weave.types.NoneType())
    assert weave.use(x) == [None, None]


def test_tagged_none():
    node = weave.save(None, "null")
    tagged_node = _test_op_tag_input(node, 1)
    assert tagged_node.type == tagged_value_type.TaggedValueType(
        weave.types.TypedDict({"a_tag": weave.types.Int()}), weave.types.NoneType()
    )
    assert weave.use(get_a_tag(tagged_node)) == 1


def test_pick_none():
    node = weave.save(None)
    assert node.type == weave.types.NoneType()
    assert weave.use(node["a"]) == None


def test_optional():
    node = weave.save([1, None])
    assert node[0].type == weave.types.optional(weave.types.Int())
    assert weave.use(node[0]) == 1
    assert node[1].type == weave.types.optional(weave.types.Int())
    assert weave.use(node[1]) == None


def test_out_of_bounds_tag_access():
    node = weave.save([1, 2])
    tagged = node._test_op_tag_input(1)
    assert weave.use(get_a_tag(tagged[0])) == 1
    assert weave.use(tagged[0]) == 1
    assert weave.use(get_a_tag(tagged[1])) == 1
    assert weave.use(tagged[1]) == 2
    assert weave.use(get_a_tag(tagged[2])) == 1
    assert weave.use(tagged[2]) == None


def test_refine_nullability():
    assert _test_op_refining(None).type == weave.types.NoneType()



import typing

from . import artifact_base
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import errors
from weave.legacy.weave import ref_base


class MemArtifact(artifact_base.Artifact):
    _refs: typing.Dict[str, "MemArtifactRef"]

    def __init__(self) -> None:
        self._refs = {}

    def ref_count(self) -> int:
        return len(self._refs)

    def refs(self) -> typing.Iterable[artifact_base.ArtifactRef]:
        return self._refs.values()

    def set(
        self, key: str, type_: types.Type, obj: typing.Any
    ) -> artifact_base.ArtifactRef:
        existing_ref = ref_base.get_ref(obj)
        if isinstance(existing_ref, artifact_base.ArtifactRef):
            return existing_ref
        ref = MemArtifactRef(self, key, type_, obj)
        self._refs[key] = ref
        return ref

    def get(self, key: str, type_: types.Type) -> typing.Any:
        return self._refs[key]._obj


class MemArtifactRef(artifact_base.ArtifactRef):
    @property
    def is_saved(self) -> bool:
        return False

    def local_ref_str(self) -> str:
        if self.path is None:
            raise errors.WeaveInternalError("MemArtifact without path!")
        return self.path

    @property
    def uri(self) -> str:
        raise errors.WeaveInternalError("Cannot get URI for in-memory artifact")



import dataclasses
import datetime
import typing

import weave
from weave.legacy.weave import op_def, ops_primitives
from weave.legacy.weave.language_features.tagging import tagged_value_type
from weave.legacy.weave.timestamp import PY_DATETIME_MAX_MS, PY_DATETIME_MIN_MS

from .concrete_tagged_value import TaggedValue


@dataclasses.dataclass
class OpSpecTestCase:
    input: tuple[typing.Any, ...]
    expected: typing.Any
    expected_type: weave.types.Type


@dataclasses.dataclass
class OpKind:
    arity: int
    uniform_input_types: bool = True
    commutative: bool = False
    list_op: bool = False
    list_list_op: bool = False


@dataclasses.dataclass
class OpSpec:
    op: op_def.OpDef
    test_cases: list[OpSpecTestCase]
    kind: OpKind


string_add = OpSpec(
    op=ops_primitives.String.__add__,
    kind=OpKind(arity=2, commutative=False),
    test_cases=[
        OpSpecTestCase(
            input=("a", "b"),
            expected="ab",
            expected_type=weave.types.String(),
        ),
        OpSpecTestCase(
            input=("b", "a"),
            expected="ba",
            expected_type=weave.types.String(),
        ),
        OpSpecTestCase(
            input=("", "b"),
            expected="b",
            expected_type=weave.types.String(),
        ),
        OpSpecTestCase(
            input=("b", ""),
            expected="b",
            expected_type=weave.types.String(),
        ),
    ],
)

number_add = OpSpec(
    op=ops_primitives.Number.__add__,
    kind=OpKind(arity=2, commutative=True),
    test_cases=[
        OpSpecTestCase(
            input=(1, 2),
            expected=3,
            expected_type=weave.types.Number(),
        ),
        OpSpecTestCase(
            input=(-1, 2),
            expected=1,
            expected_type=weave.types.Number(),
        ),
    ],
)

number_mul = OpSpec(
    op=ops_primitives.Number.__mul__,
    kind=OpKind(arity=2, commutative=True),
    test_cases=[
        OpSpecTestCase(
            input=(1, 2),
            expected=2,
            expected_type=weave.types.Number(),
        ),
        OpSpecTestCase(
            input=(-1, 2),
            expected=-2,
            expected_type=weave.types.Number(),
        ),
    ],
)

number_sub = OpSpec(
    op=ops_primitives.Number.__sub__,
    kind=OpKind(arity=2, commutative=False),
    test_cases=[
        OpSpecTestCase(
            input=(1, 2),
            expected=-1,
            expected_type=weave.types.Number(),
        ),
        OpSpecTestCase(
            input=(2, 1),
            expected=1,
            expected_type=weave.types.Number(),
        ),
        OpSpecTestCase(
            input=(-1, 2),
            expected=-3,
            expected_type=weave.types.Number(),
        ),
        OpSpecTestCase(
            input=(2, -1),
            expected=3,
            expected_type=weave.types.Number(),
        ),
    ],
)

numbers_max = OpSpec(
    op=ops_primitives.numbers_max,
    kind=OpKind(arity=1, list_op=True),
    test_cases=[
        OpSpecTestCase(
            input=([-10, -99, -100],),
            expected=-10,
            expected_type=weave.types.Number(),
        ),
        OpSpecTestCase(
            input=([0, 1, 3],),
            expected=3,
            expected_type=weave.types.Number(),
        ),
    ],
)

numbers_min = OpSpec(
    op=ops_primitives.numbers_min,
    kind=OpKind(arity=1, list_op=True),
    test_cases=[
        OpSpecTestCase(
            input=([-10, -99, -100],),
            expected=-100,
            expected_type=weave.types.Number(),
        ),
        OpSpecTestCase(
            input=([0, 1, 3],),
            expected=0,
            expected_type=weave.types.Number(),
        ),
    ],
)

numbers_avg = OpSpec(
    op=ops_primitives.numbers_avg,
    kind=OpKind(arity=1, list_op=True),
    test_cases=[
        OpSpecTestCase(
            input=([-10, -100],),
            expected=-55.0,
            expected_type=weave.types.Float(),
        ),
        OpSpecTestCase(
            input=([1, 3],),
            expected=2.0,
            expected_type=weave.types.Float(),
        ),
    ],
)

numbers_sum = OpSpec(
    op=ops_primitives.numbers_sum,
    kind=OpKind(arity=1, list_op=True),
    test_cases=[
        OpSpecTestCase(
            input=([-10, -99, -100],),
            expected=-209,
            expected_type=weave.types.Number(),
        ),
        OpSpecTestCase(
            input=([0, 1, 3],),
            expected=4,
            expected_type=weave.types.Number(),
        ),
    ],
)

index = OpSpec(
    op=ops_primitives.List.__getitem__,
    kind=OpKind(arity=2, list_op=True, uniform_input_types=False),
    test_cases=[
        OpSpecTestCase(
            input=([1, 2, 3], 0),
            expected=1,
            expected_type=weave.types.Number(),
        ),
        OpSpecTestCase(
            input=([1, 2, 3], 2),
            expected=3,
            expected_type=weave.types.Number(),
        ),
        # I didn't implement negative indexing or list indexing in vectorized
        # cases yet
        # TODO: fix!
        # OpSpecTestCase(
        #     input=([1, 2, 3], -1),
        #     expected=3,
        #     expected_type=weave.types.Number(),
        # ),
        # OpSpecTestCase(
        #     input=([1, 2, 3], [0, -1, 5, None]),
        #     expected=3,
        #     expected_type=weave.types.List(weave.types.Number()),
        # ),
        OpSpecTestCase(
            input=([-5, -6], 2),
            expected=None,
            expected_type=weave.types.optional(weave.types.Number()),
        ),
        OpSpecTestCase(
            input=([], 0),
            expected=None,
            expected_type=weave.types.NoneType(),
        ),
    ],
)

join_all = OpSpec(
    op=ops_primitives.join_all,
    kind=OpKind(arity=3, list_list_op=True, uniform_input_types=False),
    test_cases=[
        OpSpecTestCase(
            input=(
                [[{"a": 5, "b": 6}, {"a": 7, "b": 9}], [{"a": 5, "b": 8}]],
                lambda x: x["a"],
                False,
            ),
            expected=[TaggedValue({"joinObj": 5}, {"a": [5, 5], "b": [6, 8]})],
            expected_type=weave.types.List(
                tagged_value_type.TaggedValueType(
                    weave.types.TypedDict({"joinObj": weave.types.Int()}),
                    weave.types.TypedDict(
                        {
                            "a": weave.types.List(
                                weave.types.optional(weave.types.Int())
                            ),
                            "b": weave.types.List(
                                weave.types.optional(weave.types.Int())
                            ),
                        }
                    ),
                )
            ),
        ),
        # 'b' will end up a union of TypedDict and Int
        OpSpecTestCase(
            input=(
                [
                    [{"a": 5, "b": 6}, {"a": 7, "b": 9}],
                    [{"a": 5, "b": {"j": 12, "k": 13}}],
                ],
                lambda x: x["a"],
                False,
            ),
            expected=[
                TaggedValue({"joinObj": 5}, {"a": [5, 5], "b": [6, {"j": 12, "k": 13}]})
            ],
            expected_type=weave.types.List(
                tagged_value_type.TaggedValueType(
                    weave.types.TypedDict({"joinObj": weave.types.Int()}),
                    weave.types.TypedDict(
                        {
                            "a": weave.types.List(
                                weave.types.optional(weave.types.Int())
                            ),
                            "b": weave.types.List(
                                weave.types.optional(
                                    weave.types.UnionType(
                                        weave.types.Int(),
                                        weave.types.TypedDict(
                                            {
                                                "j": weave.types.Int(),
                                                "k": weave.types.Int(),
                                            }
                                        ),
                                    )
                                )
                            ),
                        }
                    ),
                )
            ),
        ),
    ],
)


dropna = OpSpec(
    op=ops_primitives.List.dropna,
    kind=OpKind(arity=1, list_op=True),
    test_cases=[
        OpSpecTestCase(
            input=([1, 2, 3],),
            expected=[1, 2, 3],
            expected_type=weave.types.List(weave.types.Int()),
        ),
        OpSpecTestCase(
            input=([1, None, 3],),
            expected=[1, 3],
            expected_type=weave.types.List(weave.types.Int()),
        ),
        OpSpecTestCase(
            input=([None, None, None],),
            expected=[],
            expected_type=weave.types.List(weave.types.NoneType()),
        ),
    ],
)

number_to_timestamp = OpSpec(
    op=ops_primitives.Number.to_timestamp,
    kind=OpKind(arity=1),
    test_cases=[
        # Current Time
        OpSpecTestCase(
            input=(1677098489000,),
            expected=datetime.datetime.fromtimestamp(1677098489, datetime.timezone.utc),
            expected_type=weave.types.Timestamp(),
        ),
        # Zero Time
        OpSpecTestCase(
            input=(0,),
            expected=datetime.datetime.fromtimestamp(0, datetime.timezone.utc),
            expected_type=weave.types.Timestamp(),
        ),
        # Negative Time
        OpSpecTestCase(
            input=(-1677098489000,),
            expected=datetime.datetime.fromtimestamp(
                -1677098489, datetime.timezone.utc
            ),
            expected_type=weave.types.Timestamp(),
        ),
        # Small Value
        OpSpecTestCase(
            input=(1000,),
            expected=datetime.datetime.fromtimestamp(1, datetime.timezone.utc),
            expected_type=weave.types.Timestamp(),
        ),
        # Large Value
        OpSpecTestCase(
            input=(PY_DATETIME_MAX_MS,),
            expected=datetime.datetime.fromtimestamp(
                PY_DATETIME_MAX_MS / 1000, datetime.timezone.utc
            ),
            expected_type=weave.types.Timestamp(),
        ),
        # Large Value (in nanoseconds)
        OpSpecTestCase(
            input=(PY_DATETIME_MAX_MS + 12345,),
            expected=datetime.datetime.fromtimestamp(
                ((PY_DATETIME_MAX_MS + 12345) // 1000) / 1000, datetime.timezone.utc
            ),
            expected_type=weave.types.Timestamp(),
        ),
        # Large Negative Value
        OpSpecTestCase(
            input=(PY_DATETIME_MIN_MS,),
            expected=datetime.datetime.fromtimestamp(
                PY_DATETIME_MIN_MS / 1000, datetime.timezone.utc
            ),
            expected_type=weave.types.Timestamp(),
        ),
        # Large Negative Value (in nano seconds)
        OpSpecTestCase(
            input=(PY_DATETIME_MIN_MS - 12345,),
            expected=datetime.datetime.fromtimestamp(
                ((PY_DATETIME_MIN_MS - 12345) // 1000) / 1000, datetime.timezone.utc
            ),
            expected_type=weave.types.Timestamp(),
        ),
        # Float Value (Drops Fractional Part)
        OpSpecTestCase(
            input=(1677098489215.025,),
            expected=datetime.datetime.fromtimestamp(
                1677098489.215, datetime.timezone.utc
            ),
            expected_type=weave.types.Timestamp(),
        ),
    ],
)

number_equal = OpSpec(
    op=ops_primitives.Number.__eq__,
    kind=OpKind(arity=2, commutative=True),
    test_cases=[
        OpSpecTestCase(
            input=(1, 1),
            expected=True,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(1, 2),
            expected=False,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(1, None),
            expected=False,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(None, None),
            expected=True,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(None, 1),
            expected=False,
            expected_type=weave.types.Boolean(),
        ),
    ],
)


string_equal = OpSpec(
    op=ops_primitives.String.__eq__,
    kind=OpKind(arity=2, commutative=True),
    test_cases=[
        OpSpecTestCase(
            input=("a", "a"),
            expected=True,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=("a", "b"),
            expected=False,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=("a", None),
            expected=False,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(None, None),
            expected=True,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(None, "a"),
            expected=False,
            expected_type=weave.types.Boolean(),
        ),
    ],
)


boolean_equal = OpSpec(
    op=ops_primitives.Boolean.bool_equals,
    kind=OpKind(arity=2, commutative=True),
    test_cases=[
        OpSpecTestCase(
            input=(True, True),
            expected=True,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(True, False),
            expected=False,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(True, None),
            expected=False,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(None, None),
            expected=True,
            expected_type=weave.types.Boolean(),
        ),
        OpSpecTestCase(
            input=(None, True),
            expected=False,
            expected_type=weave.types.Boolean(),
        ),
    ],
)


OP_TEST_SPECS = [
    number_add,
    number_sub,
    number_mul,
    string_add,
    numbers_max,
    numbers_min,
    numbers_avg,
    numbers_sum,
    number_to_timestamp,
    index,
    join_all,
    dropna,
    number_equal,
    string_equal,
    boolean_equal,
]



import math

from weave.legacy.weave import api, context, mappers_python, val_const, weave_internal
from weave.legacy.weave import weave_types as types


def test_map_typed_dict():
    d = {"a": 5, "b": "x"}
    d_type = types.TypeRegistry.type_of(d)
    m = mappers_python.map_to_python(d_type, None)
    # TODO: This fails because we start with ConstString and end up with String
    # assert d_type == d2_type
    d2 = m.apply(d)
    assert d == d2
    m2 = mappers_python.map_from_python(d_type, None)
    d3 = m2.apply(d2)
    assert d2 == d3


def test_map_dict():
    d = {"a": 5, "b": 9}
    d_type = types.Dict(types.String(), types.Int())
    m = mappers_python.map_to_python(d_type, None)
    d2 = m.apply(d)
    assert d == d2
    m2 = mappers_python.map_from_python(d_type, None)
    d3 = m2.apply(d2)
    assert d2 == d3


def test_map_type():
    t = types.TypedDict({"a": types.Int(), "b": types.String()})
    t_type = types.TypeRegistry.type_of(t)
    m = mappers_python.map_to_python(t_type, None)
    t2 = m.apply(t)
    m2 = mappers_python.map_from_python(t_type, None)
    t3 = m2.apply(t2)
    assert t == t3


def test_map_typed_dict_with_nan():
    d = {"a": 5, "b": float("nan")}
    d_type = types.TypeRegistry.type_of(d)
    m = mappers_python.map_to_python(d_type, None)
    d2 = m.apply(d)
    m2 = mappers_python.map_from_python(d_type, None)
    d3 = m2.apply(d2)
    assert d["a"] == d3["a"]
    assert math.isnan(d3["b"])


def test_map_const():
    d = {"a": val_const.const(5)}
    d_type = types.TypeRegistry.type_of(d)
    m = mappers_python.map_to_python(d_type, None)
    d2 = m.apply(d)
    m2 = mappers_python.map_from_python(d_type, None)
    d3 = m2.apply(d2)
    # const is the only case where the mapping is not two-way.
    # TODO: is this a problem?
    assert d3["a"] == 5


def test_map_union_none_type():
    d = None
    d_type = types.UnionType(types.TypeType(), types.NoneType())
    m = mappers_python.map_to_python(d_type, None)
    d2 = m.apply(d)
    m2 = mappers_python.map_from_python(d_type, None)
    d3 = m2.apply(d2)
    assert d3 is None



import weave


def inheritors(klass):
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses


def all_panels():
    return inheritors(weave.legacy.weave.panel.Panel)


def test_panel_coverage():
    panels = all_panels()
    panel_ids = set([p.id for p in panels])
    py_panels = set(panel_ids)
    missing_panels = list(js_registered_panels - py_panels)
    total_missing = len(missing_panels)
    error_msg = f"**{total_missing} Missing Panels**\n\n" + "\n".join(missing_panels)

    assert total_missing == 0, error_msg


js_registered_panels = set(
    [
        "Expression",
        "id-compare",
        "id-compare-count",
        "id-count",
        "type",
        "date",
        "boolean",
        "table",
        "plot",
        "Facet",
        "number",
        "barchart",
        "histogram",
        "precomputed-histogram",
        "multi-histogram",
        "string",
        "string-compare",
        "string-histogram",
        "multi-string-histogram",
        "none",
        "unknown",
        "link",
        "object",
        "project-overview",
        "run-overview",
        "run-color",
        "image-file",
        "image-file-compare",
        "jupyter",
        "rawimage",
        "markdown",
        "textdiff",
        "text",
        "netron",
        "tracer",
        "web-viz",
        "video-file",
        "audio-file",
        "html-file",
        "bokeh-file",
        "object3D-file",
        "molecule-file",
        "model-file",
        "dir",
        "ndarray",
        "artifactVersionAliases",
        "wb_trace_tree-traceDebugger",
        "wb_trace_tree-traceViewer",
        "wb_trace_tree-modelViewer",
        "debug-expression-graph",
    ]
)



import importlib
from typing import TYPE_CHECKING, Callable, Dict, Optional

if TYPE_CHECKING:
    from groq.types.chat import ChatCompletion, ChatCompletionChunk

import weave
from weave.trace.op_extensions.accumulator import add_accumulator
from weave.trace.patcher import MultiPatcher, SymbolPatcher


def groq_accumulator(
    acc: Optional["ChatCompletion"], value: "ChatCompletionChunk"
) -> "ChatCompletion":
    from groq.types.chat import ChatCompletion, ChatCompletionMessage
    from groq.types.chat.chat_completion import Choice
    from groq.types.chat.chat_completion_chunk import Choice as ChoiceChunk
    from groq.types.completion_usage import CompletionUsage

    if acc is None:
        choices = []
        for choice in value.choices:
            if isinstance(choice, ChoiceChunk):
                choices.append(
                    Choice(
                        message=ChatCompletionMessage(
                            content=choice.delta.content,
                            role=choice.delta.role,
                            function_call=choice.delta.function_call,
                            tool_call=choice.delta.tool_calls,
                        ),
                        finish_reason="stop",
                        index=choice.index,
                        logprobs=choice.logprobs,
                    )
                )
            else:
                choices.append(choice)
        acc = ChatCompletion(
            id=value.id,
            choices=choices,
            created=value.created,
            model=value.model,
            object="chat.completion",
            system_fingerprint=value.system_fingerprint,
            usage=CompletionUsage(
                completion_tokens=0,
                prompt_tokens=0,
                total_tokens=0,
                completion_time=0,
                prompt_time=0,
                queue_time=0,
                total_time=0,
            ),
        )

    if value.x_groq is not None and value.x_groq.usage is not None:
        acc.usage.completion_tokens += value.x_groq.usage.completion_tokens
        acc.usage.prompt_tokens += value.x_groq.usage.prompt_tokens
        acc.usage.total_tokens += value.x_groq.usage.total_tokens
        acc.usage.completion_time += value.x_groq.usage.completion_time
        acc.usage.prompt_time += value.x_groq.usage.prompt_time
        acc.usage.queue_time += value.x_groq.usage.queue_time
        acc.usage.total_time += value.x_groq.usage.total_time

    if value.choices:
        for idx, choice in enumerate(value.choices):
            if isinstance(acc.choices[idx].message.content, str) and isinstance(
                choice.delta.content, str
            ):
                acc.choices[idx].message.content += choice.delta.content
            if acc.choices[idx].message.function_call:
                acc.choices[idx].message.function_call.append(
                    choice.delta.function_call
                )
            if acc.choices[idx].message.tool_call:
                acc.choices[idx].message.tool_call.append(choice.delta.tool_call)

    return acc


def should_use_accumulator(inputs: Dict) -> bool:
    return isinstance(inputs, dict) and bool(inputs.get("stream"))


def groq_wrapper(name: str) -> Callable[[Callable], Callable]:
    def wrapper(fn: Callable) -> Callable:
        op = weave.op()(fn)
        op.name = name  # type: ignore
        # return op
        return add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda inputs: groq_accumulator,
            should_accumulate=should_use_accumulator,
        )

    return wrapper


groq_patcher = MultiPatcher(
    [
        SymbolPatcher(
            lambda: importlib.import_module("groq.resources.chat.completions"),
            "Completions.create",
            groq_wrapper(name="groq.chat.completions.create"),
        ),
        SymbolPatcher(
            lambda: importlib.import_module("groq.resources.chat.completions"),
            "AsyncCompletions.create",
            groq_wrapper(name="groq.async.chat.completions.create"),
        ),
    ]
)



import replicate
import os
import tempfile
import weave

from PIL import Image

import requests
import shutil
import logging

logger = logging.getLogger(__name__)

from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)


try:
    with open(os.path.expanduser("~/.replicate_api_token")) as f:
        os.environ["REPLICATE_API_TOKEN"] = f.read().strip()
except FileNotFoundError:
    pass

retry_replicate_decorator = retry(
    reraise=True,
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    retry=(retry_if_exception_type(replicate.exceptions.ModelError)),
    before_sleep=before_sleep_log(logger, logging.WARNING),
)


def download_file(url, local_path):
    r = requests.get(url, stream=True)
    if r.status_code == 200:
        with open(local_path, "wb") as f:
            r.raw.decode_content = True
            shutil.copyfileobj(r.raw, f)


@weave.op()
@retry_replicate_decorator
def stable_diffusion(prompt: str) -> Image.Image:
    print("Running stable diffusion", prompt)
    model = replicate.models.get("stability-ai/stable-diffusion")
    image_url = model.predict(prompt=prompt)[0]
    image_file = tempfile.NamedTemporaryFile()
    download_file(image_url, image_file.name)
    image_file.seek(0)
    return Image.open(image_file)


def replicate_image_to_text(image: Image.Image, model_id: str) -> str:
    print("Running %s: %s" % (model_id, image))
    model = replicate.models.get(model_id)
    # save image to a temporary directory and upload it to Replicate
    with tempfile.TemporaryDirectory() as d:
        image.save(os.path.join(d, "image.png"))
        result = model.predict(image=open(os.path.join(d, "image.png"), "rb"))
    return result


@weave.op()
@retry_replicate_decorator
def img2prompt(image: Image.Image) -> str:
    return replicate_image_to_text(image, "methexis-inc/img2prompt")


@weave.op()
@retry_replicate_decorator
def clip_prefix_caption(image: Image.Image) -> str:
    return replicate_image_to_text(image, "rmokady/clip_prefix_caption")



import logging
from typing import Any, Callable, Optional, Sequence

from weave.trace.context import get_raise_on_captured_errors

logger = logging.getLogger(__name__)


class Patcher:
    def attempt_patch(self) -> bool:
        raise NotImplementedError()

    def undo_patch(self) -> bool:
        raise NotImplementedError()


class MultiPatcher(Patcher):
    def __init__(self, patchers: Sequence[Patcher]) -> None:
        self.patchers = patchers

    def attempt_patch(self) -> bool:
        all_successful = True
        for patcher in self.patchers:
            try:
                all_successful = all_successful and patcher.attempt_patch()
            except Exception as e:
                logger.error(f"Error patching - some logs may not be captured: {e}")
                if get_raise_on_captured_errors():
                    raise
                all_successful = False
        return all_successful

    def undo_patch(self) -> bool:
        all_successful = True
        for patcher in self.patchers:
            try:
                all_successful = all_successful and patcher.undo_patch()
            except Exception as e:
                logger.error(f"Error unpatching: {e}")
                if get_raise_on_captured_errors():
                    raise
                all_successful = False
        return all_successful


class _SymbolTarget:
    def __init__(self, base_symbol: Any, attr: str) -> None:
        self.base_symbol = base_symbol
        self.attr = attr


class SymbolPatcher(Patcher):
    _get_base_symbol: Callable
    _attribute_name: str
    _make_new_value: Callable
    _original_value: Any = None

    def __init__(
        self, get_base_symbol: Callable, attribute_name: str, make_new_value: Callable
    ) -> None:
        self._get_base_symbol = get_base_symbol
        self._attribute_name = attribute_name
        self._make_new_value = make_new_value

    def _get_symbol_target(self) -> Optional[_SymbolTarget]:
        try:
            base_symbol = self._get_base_symbol()
        except Exception:
            return None
        if base_symbol is None:
            return None
        parts = self._attribute_name.split(".")
        for part in parts[:-1]:
            try:
                base_symbol = getattr(base_symbol, part)
            except AttributeError:
                return None
        attr = parts[-1]
        return _SymbolTarget(base_symbol, attr)

    def attempt_patch(self) -> bool:
        if self._original_value:
            return True
        target = self._get_symbol_target()
        if target is None:
            return False
        original_value = getattr(target.base_symbol, target.attr)
        try:
            new_val = self._make_new_value(original_value)
        except Exception:
            logger.error(f"Failed to patch {self._attribute_name}")
            return False
        setattr(
            target.base_symbol,
            target.attr,
            new_val,
        )
        self._original_value = original_value
        return True

    def undo_patch(self) -> bool:
        if not self._original_value:
            return False
        target = self._get_symbol_target()
        if target is None:
            return False

        setattr(target.base_symbol, target.attr, self._original_value)
        self._original_value = None
        return True



import typing

import weave


def test_type_var_in_dict_any():
    @weave.type()
    class TestTypeVarsInDictType:
        v: dict[str, typing.Any]

    assert len(TestTypeVarsInDictType.WeaveType.type_attrs()) == 1


def test_object_union_attr_is_variable():
    @weave.type()
    class ObjWithUnion:
        a: typing.Union[str, int]

    assert "a" in ObjWithUnion.WeaveType().type_vars


def test_object_noneunion_attr_is_variable():
    @weave.type()
    class ObjWithUnion:
        a: weave.Node[typing.Union[str, int]]

    assert "a" in ObjWithUnion.WeaveType().type_vars


def test_type_is_reloctable():
    @weave.type()
    class CoolObjBase:
        pass

    @weave.type()
    class CoolObj(CoolObjBase):
        a: int
        b: str

    obj = CoolObj(1, "hi")
    ref = weave.storage.save(obj)
    obj2 = weave.storage.get(str(ref))
    assert obj2.a == 1


def test_type_with_name():
    @weave.type()
    class CoolObjWithName:
        name: str
        a: int

    obj = CoolObjWithName("my-name", 15)
    ref = weave.storage.save(obj)
    obj2 = weave.storage.get(str(ref))
    assert obj2.a == 15
    assert obj2.name == "my-name"



# These are performance tests. If they start to flake we can come up
# with a better way to run them.
import time

import pytest

import weave
from weave.legacy.weave import ops_arrow, ops_primitives


@pytest.mark.skip(reason="Performance test")
def test_listmap():
    range_data = list(range(20000))
    pydata = [{"a": i, "b": i, "c": i} for i in range_data]
    data = ops_arrow.to_weave_arrow(pydata)
    node = ops_primitives.List._listmap(data, lambda x: x["a"] + 1)
    expected = [x + 1 for x in range_data]
    start_time = time.time()
    assert weave.use(node) == expected
    elapsed = time.time() - start_time
    # Runs in 0.9s on my m1 macbook pro
    assert elapsed < 2.5



import importlib
import typing
from functools import wraps

import weave
from weave.trace.patcher import MultiPatcher, SymbolPatcher


def create_wrapper_sync(
    name: str,
) -> typing.Callable[[typing.Callable], typing.Callable]:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        op = weave.op()(fn)
        op.name = name  # type: ignore
        return op

    return wrapper


def create_wrapper_async(
    name: str,
) -> typing.Callable[[typing.Callable], typing.Callable]:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        def _fn_wrapper(fn: typing.Callable) -> typing.Callable:
            @wraps(fn)
            async def _async_wrapper(
                *args: typing.Any, **kwargs: typing.Any
            ) -> typing.Any:
                return await fn(*args, **kwargs)

            return _async_wrapper

        op = weave.op()(_fn_wrapper(fn))
        op.name = name  # type: ignore
        return op

    return wrapper


cerebras_patcher = MultiPatcher(
    [
        SymbolPatcher(
            lambda: importlib.import_module("cerebras.cloud.sdk.resources.chat"),
            "CompletionsResource.create",
            create_wrapper_sync(name="cerebras.chat.completions.create"),
        ),
        SymbolPatcher(
            lambda: importlib.import_module("cerebras.cloud.sdk.resources.chat"),
            "AsyncCompletionsResource.create",
            create_wrapper_async(name="cerebras.chat.completions.create"),
        ),
    ]
)



import pytest

import weave
from weave.flow.obj import Object
from weave.legacy.weave import artifact_local, storage
from weave.legacy.weave import ops_arrow as arrow
from weave.trace import ref_util
from weave.trace_server.refs_internal import (
    DICT_KEY_EDGE_NAME,
    LIST_INDEX_EDGE_NAME,
    OBJECT_ATTR_EDGE_NAME,
)


def test_laref_artifact_version_1():
    art = artifact_local.LocalArtifact("my-art", "v19")
    ref = artifact_local.LocalArtifactRef(art, "obj")
    assert str(ref) == "local-artifact:///my-art:v19/obj"

    print("STR", str(ref))
    parsed_ref = artifact_local.LocalArtifactRef.from_str(str(ref))
    assert parsed_ref.artifact.name == art.name
    assert parsed_ref.artifact._version == art._version
    assert parsed_ref.path == "obj"
    assert parsed_ref.extra is None

    assert ref.local_ref_str() == "obj"
    assert ref_util.parse_local_ref_str("obj") == ("obj", None)


def test_laref_artifact_version_path():
    art = artifact_local.LocalArtifact("my-art", "v19")
    ref = artifact_local.LocalArtifactRef(art, "x.txt")
    assert str(ref) == "local-artifact:///my-art:v19/x.txt"

    parsed_ref = artifact_local.LocalArtifactRef.from_str(str(ref))
    assert parsed_ref.artifact.name == art.name
    assert parsed_ref.artifact._version == art._version
    assert parsed_ref.path == "x.txt"
    assert parsed_ref.extra is None

    assert ref.local_ref_str() == "x.txt"
    assert ref_util.parse_local_ref_str("x.txt") == (
        "x.txt",
        None,
    )


def test_laref_artifact_version_path_extra1():
    art = artifact_local.LocalArtifact("my-art", "v19")
    ref = artifact_local.LocalArtifactRef(art, "x.txt", extra=["5"])
    assert str(ref) == "local-artifact:///my-art:v19/x.txt#5"

    parsed_ref = artifact_local.LocalArtifactRef.from_str(str(ref))
    assert parsed_ref.artifact.name == art.name
    assert parsed_ref.artifact._version == art._version
    assert parsed_ref.path == "x.txt"
    assert parsed_ref.extra == ["5"]

    assert ref.local_ref_str() == "x.txt#5"
    assert ref_util.parse_local_ref_str("x.txt#5") == (
        "x.txt",
        ["5"],
    )


def test_laref_artifact_version_path_obj_extra1():
    art = artifact_local.LocalArtifact("my-art", "v19")
    ref = artifact_local.LocalArtifactRef(art, "obj", extra=["5"])
    assert str(ref) == "local-artifact:///my-art:v19/obj#5"

    parsed_ref = artifact_local.LocalArtifactRef.from_str(str(ref))
    assert parsed_ref.artifact.name == art.name
    assert parsed_ref.artifact._version == art._version
    assert parsed_ref.path == "obj"
    assert parsed_ref.extra == ["5"]

    assert ref.local_ref_str() == "obj#5"
    assert ref_util.parse_local_ref_str("obj#5") == (
        "obj",
        ["5"],
    )


def test_laref_artifact_version_path_extra2():
    art = artifact_local.LocalArtifact("my-art", "v19")
    ref = artifact_local.LocalArtifactRef(art, "x.txt", extra=["5", "a"])
    assert str(ref) == "local-artifact:///my-art:v19/x.txt#5/a"

    parsed_ref = artifact_local.LocalArtifactRef.from_str(str(ref))
    assert parsed_ref.artifact.name == art.name
    assert parsed_ref.artifact._version == art._version
    assert parsed_ref.path == "x.txt"
    assert parsed_ref.extra == ["5", "a"]

    assert ref.local_ref_str() == "x.txt#5/a"
    assert ref_util.parse_local_ref_str("x.txt#5/a") == (
        "x.txt",
        ["5", "a"],
    )


"""

* Dict
    * Key
* List
    * Index
* Object
    * Attribute
* Table (ArrowWeaveList)
    * Index
    * Column
    * Id (not implemented)

"""


def assert_local_ref(object, path_parts, extra_parts):
    obj_ref = storage.get_ref(object)
    parsed_obj_ref = ref_util.parse_ref_str(obj_ref.uri)

    target_obj_ref = ref_util.ParsedRef(
        scheme="local-artifact",
        entity=None,
        project=None,
        artifact=obj_ref.name,
        alias=obj_ref.version,
        file_path_parts=path_parts,
        ref_extra_tuples=[
            ref_util.RefExtraTuple(*extra_parts[i : i + 2])
            for i in range(0, len(extra_parts), 2)
        ],
    )
    assert parsed_obj_ref == target_obj_ref


def test_ref_parser():
    assert ref_util.parse_ref_str(
        "local-artifact:///art_name:version"
    ) == ref_util.ParsedRef(
        scheme="local-artifact",
        entity=None,
        project=None,
        artifact="art_name",
        alias="version",
        file_path_parts=[],
        ref_extra_tuples=[],
    )
    assert ref_util.parse_ref_str(
        "local-artifact:///art_name:version/possibly/deep/path"
    ) == ref_util.ParsedRef(
        scheme="local-artifact",
        entity=None,
        project=None,
        artifact="art_name",
        alias="version",
        file_path_parts=["possibly", "deep", "path"],
        ref_extra_tuples=[],
    )
    assert ref_util.parse_ref_str(
        "local-artifact:///art_name:version/possibly/deep/path#very/deep/ref/extra"
    ) == ref_util.ParsedRef(
        scheme="local-artifact",
        entity=None,
        project=None,
        artifact="art_name",
        alias="version",
        file_path_parts=["possibly", "deep", "path"],
        ref_extra_tuples=[
            ref_util.RefExtraTuple("very", "deep"),
            ref_util.RefExtraTuple("ref", "extra"),
        ],
    )
    assert ref_util.parse_ref_str(
        "wandb-artifact:///entity/project/art_name:version"
    ) == ref_util.ParsedRef(
        scheme="wandb-artifact",
        entity="entity",
        project="project",
        artifact="art_name",
        alias="version",
        file_path_parts=[],
        ref_extra_tuples=[],
    )
    assert ref_util.parse_ref_str(
        "wandb-artifact:///entity/project/art_name:version/possibly/deep/path"
    ) == ref_util.ParsedRef(
        scheme="wandb-artifact",
        entity="entity",
        project="project",
        artifact="art_name",
        alias="version",
        file_path_parts=["possibly", "deep", "path"],
        ref_extra_tuples=[],
    )
    assert ref_util.parse_ref_str(
        "wandb-artifact:///entity/project/art_name:version/possibly/deep/path#very/deep/ref/extra"
    ) == ref_util.ParsedRef(
        scheme="wandb-artifact",
        entity="entity",
        project="project",
        artifact="art_name",
        alias="version",
        file_path_parts=["possibly", "deep", "path"],
        ref_extra_tuples=[
            ref_util.RefExtraTuple("very", "deep"),
            ref_util.RefExtraTuple("ref", "extra"),
        ],
    )


def test_ref_extra_dict(ref_tracking):
    obj = {"a": 1}
    saved_obj = weave.use(weave.save(obj))

    assert saved_obj == obj
    assert_local_ref(saved_obj, ["obj"], [])
    assert storage.get(saved_obj._ref.uri) == obj

    val = saved_obj["a"]
    assert val == 1
    assert_local_ref(val, ["obj"], [DICT_KEY_EDGE_NAME, "a"])
    assert storage.get(val._ref.uri) == 1


def test_ref_extra_list(ref_tracking):
    obj = [1]
    saved_obj = weave.use(weave.save(obj))

    assert saved_obj == obj
    assert_local_ref(saved_obj, ["obj"], [])
    assert storage.get(saved_obj._ref.uri) == obj

    val = saved_obj[0]
    assert val == 1
    assert_local_ref(val, ["obj"], [LIST_INDEX_EDGE_NAME, "0"])
    assert storage.get(val._ref.uri) == 1


def make_custom_object_classes_classic_type():
    @weave.type()
    class CustomObjectA:
        inner_a: int

    @weave.type()
    class CustomObjectB:
        inner_b: CustomObjectA

    return CustomObjectA, CustomObjectB


def make_custom_object_classes_pydantic():
    class CustomObjectA(Object):
        inner_a: int

    class CustomObjectB(Object):
        inner_b: CustomObjectA

    return CustomObjectA, CustomObjectB


@pytest.mark.parametrize(
    "get_custom_object_classes",
    [make_custom_object_classes_classic_type, make_custom_object_classes_pydantic],
)
def test_ref_extra_object(ref_tracking, get_custom_object_classes):
    CustomObjectA, CustomObjectB = get_custom_object_classes()

    obj = CustomObjectA(inner_a=1)
    saved_obj = weave.use(weave.save(obj))

    assert_local_ref(saved_obj, ["obj"], [])

    val = saved_obj.inner_a
    assert val == 1
    assert_local_ref(val, ["obj"], [OBJECT_ATTR_EDGE_NAME, "inner_a"])
    assert storage.get(val._ref.uri) == 1


def test_ref_extra_table(ref_tracking):
    arrow_raw = arrow.to_arrow(
        [
            {"a": 1},
            {"a": 2},
        ]
    )
    saved_obj = weave.use(weave.save(arrow_raw))

    assert saved_obj.to_pylist_notags() == arrow_raw.to_pylist_notags()
    assert_local_ref(saved_obj, ["obj"], [])
    assert (
        storage.get(saved_obj._ref.uri).to_pylist_notags()
        == arrow_raw.to_pylist_notags()
    )

    val = saved_obj[0]
    assert val == {"a": 1}
    assert_local_ref(val, ["obj"], ["row", "0"])
    assert storage.get(val._ref.uri) == {"a": 1}

    val = saved_obj.column("a")
    assert val.to_pylist_notags() == [1, 2]
    assert_local_ref(val, ["obj"], ["col", "a"])
    assert storage.get(val._ref.uri).to_pylist_notags() == [1, 2]


@pytest.mark.parametrize(
    "get_custom_object_classes",
    [make_custom_object_classes_classic_type, make_custom_object_classes_pydantic],
)
def test_ref_extra_table_very_nested(ref_tracking, get_custom_object_classes):
    CustomObjectA, CustomObjectB = get_custom_object_classes()

    arrow_raw = arrow.to_arrow(
        [
            {
                "a": arrow.to_arrow(
                    [
                        [CustomObjectA(inner_a=1)],
                    ]
                )
            },
        ]
    )

    saved_obj = weave.use(weave.save(arrow_raw))

    assert saved_obj.to_pylist_notags() == arrow_raw.to_pylist_notags()

    val = saved_obj[0]["a"][0][0].inner_a
    assert val == 1
    assert_local_ref(
        val,
        ["obj"],
        [
            "row",
            "0",
            DICT_KEY_EDGE_NAME,
            "a",
            "row",
            "0",
            LIST_INDEX_EDGE_NAME,
            "0",
            OBJECT_ATTR_EDGE_NAME,
            "inner_a",
        ],
    )
    assert storage.get(val._ref.uri) == 1


def test_refs_across_artifacts(ref_tracking):
    inner_obj = {"a": 1}
    saved_inner_obj = weave.use(weave.save(inner_obj))
    outer_obj = {"outer": inner_obj, "inner": saved_inner_obj}
    saved_outer_obj = weave.use(weave.save(outer_obj))
    obj = {"outer": inner_obj, "inner": {"a": 1}}

    # TODO: These need to be uncommented as they highlight an issue with nested refs
    # assert saved_outer_obj == obj
    assert_local_ref(saved_outer_obj, ["obj"], [])
    # assert storage.get(saved_outer_obj._ref.uri) == obj

    # Non-saved children follow the same ref structure
    val = saved_outer_obj["outer"]
    assert val == inner_obj
    assert_local_ref(val, ["obj"], [DICT_KEY_EDGE_NAME, "outer"])
    assert storage.get(val._ref.uri) == inner_obj

    val = saved_outer_obj["outer"]["a"]
    assert val == 1
    assert_local_ref(
        val, ["obj"], [DICT_KEY_EDGE_NAME, "outer", DICT_KEY_EDGE_NAME, "a"]
    )
    assert storage.get(val._ref.uri) == 1

    # Jumping artifact boundaries uses new artifact refs
    val = saved_outer_obj["inner"]
    # assert val == inner_obj
    assert_local_ref(val, ["obj"], [])
    assert storage.get(val._ref.uri) == inner_obj

    val = saved_outer_obj["inner"]["a"]
    # assert val == inner_obj
    assert_local_ref(val, ["obj"], [DICT_KEY_EDGE_NAME, "a"])
    assert storage.get(val._ref.uri) == 1


@pytest.mark.parametrize(
    "get_custom_object_classes",
    [make_custom_object_classes_classic_type, make_custom_object_classes_pydantic],
)
def test_ref_objects_across_artifacts_nocross(ref_tracking, get_custom_object_classes):
    CustomObjectA, CustomObjectB = get_custom_object_classes()

    # Case 1: No Cross Ref
    obj_a = CustomObjectA(inner_a=1)
    obj_b = CustomObjectB(inner_b=obj_a)
    saved_obj_b = weave.use(weave.save(obj_b))

    val = saved_obj_b.inner_b
    assert_local_ref(val, ["obj"], [OBJECT_ATTR_EDGE_NAME, "inner_b"])


@pytest.mark.parametrize(
    "get_custom_object_classes",
    [make_custom_object_classes_classic_type],
)
def test_ref_objects_across_artifacts_cross(ref_tracking, get_custom_object_classes):
    CustomObjectA, CustomObjectB = get_custom_object_classes()

    # Case 2: Cross Ref
    obj_a = CustomObjectA(inner_a=1)
    obj_a_ref = weave.use(weave.save(obj_a, "COB"))
    obj_b = CustomObjectB(inner_b=obj_a_ref)
    weave.types.type_of_with_refs(obj_b.inner_b)
    saved_obj_b = weave.use(weave.save(obj_b))

    val = saved_obj_b.inner_b
    # Notice that we have reset the ref structure here to the new artifact
    assert_local_ref(val, ["obj"], [])



"""
Contains the TemplateRegistry class, which is used to register
templates for the PyBoard generator. I think we might want to make
this even more generic and allow any panel to be a generator, but for
now this is a simple abstraction that will work for basic use cases.

# TODO: Generalize this to work with panels like /weave/ecosystem/wandb/panel_time_series.py
# Proper panels have the following advantageous patterns:
1. They can create default configs from inputs
2. They can declare their own config editor rendering function
3. Their render function is pre-defined
4. If we make it work with the panel pattern, presumably it will be more general
   and lend itself to ui-driven generation of panels.

However, this current implementation is simple and easy to refactor.
"""

import dataclasses
import typing

from weave.legacy.weave import weave_types
from weave.legacy.weave import decorator_op, graph, registry_mem


@dataclasses.dataclass
class TemplateRegistrySpec:
    display_name: str
    description: str
    op_name: str
    input_node_predicate: typing.Optional[
        typing.Callable[[graph.Node[typing.Any]], bool]
    ] = None
    config_type: typing.Optional[weave_types.Type] = None
    is_featured: typing.Optional[bool] = None
    instructions_md: typing.Optional[str] = None
    thumbnail_url: typing.Optional[str] = None


@dataclasses.dataclass
class _TemplateRegistry:
    _specs: typing.Dict[str, TemplateRegistrySpec] = dataclasses.field(
        default_factory=dict
    )

    def register(
        self,
        op_name: str,
        display_name: str,
        description: str,
        *,
        input_node_predicate: typing.Optional[
            typing.Callable[[graph.Node[typing.Any]], bool]
        ] = None,
        config_type: typing.Optional[weave_types.Type] = None,
        is_featured: typing.Optional[bool] = None,
        instructions_md: typing.Optional[str] = None,
        thumbnail_url: typing.Optional[str] = None,
    ) -> None:
        return self.register_spec(
            TemplateRegistrySpec(
                display_name=display_name,
                description=description,
                op_name=op_name,
                input_node_predicate=input_node_predicate,
                config_type=config_type,
                is_featured=is_featured,
                instructions_md=instructions_md,
                thumbnail_url=thumbnail_url,
            )
        )

    def register_spec(self, spec: TemplateRegistrySpec) -> None:
        if spec.op_name in self._specs:
            raise ValueError(f"Template {spec.op_name} already registered")
        self._specs[spec.op_name] = spec

    def get_spec(self, name: str) -> TemplateRegistrySpec:
        return self._specs[name]

    def get_specs(self) -> typing.Dict[str, TemplateRegistrySpec]:
        return {**self._specs}


class PyBoardGeneratorSpec(typing.TypedDict):
    display_name: str
    description: str
    op_name: str
    config_type: typing.Optional[weave_types.Type]
    instructions_md: typing.Optional[str]
    thumbnail_url: typing.Optional[str]


# Processes have a singleton TemplateRegistry
template_registry = _TemplateRegistry()


@decorator_op.op(name="py_board-get_board_templates_for_node", hidden=True)  # type: ignore
def get_board_templates_for_node(
    input_node: graph.Node[typing.Any],
) -> list[PyBoardGeneratorSpec]:
    final_specs = []
    for op_name, spec in template_registry.get_specs().items():
        op = registry_mem.memory_registry._ops[op_name]
        if op.input_type.first_param_valid(input_node.type):
            predicate = spec.input_node_predicate
            if predicate is None or predicate(input_node):
                final_specs.append(
                    PyBoardGeneratorSpec(
                        display_name=spec.display_name,
                        description=spec.description,
                        op_name=spec.op_name,
                        config_type=spec.config_type,
                        instructions_md=spec.instructions_md,
                        thumbnail_url=spec.thumbnail_url,
                    )
                )
    return final_specs


@decorator_op.op(name="py_board-get_featured_board_templates", hidden=True)  # type: ignore
def get_featured_board_templates() -> list[PyBoardGeneratorSpec]:
    final_specs = []
    for op_name, spec in template_registry.get_specs().items():
        if spec.is_featured and spec.instructions_md:
            final_specs.append(
                PyBoardGeneratorSpec(
                    display_name=spec.display_name,
                    description=spec.description,
                    op_name=spec.op_name,
                    config_type=spec.config_type,
                    instructions_md=spec.instructions_md,
                    thumbnail_url=spec.thumbnail_url,
                )
            )
    return final_specs



import weave
from weave.legacy.weave import panel


@weave.type()
class StringEditor(panel.Panel):
    id = "StringEditor"

    @weave.op()
    def value(self) -> str:
        return weave.use(self.input_node)



import typing

import weave
from weave.legacy.weave.ecosystem.slack import slack


@weave.type()
class SlackMessagesPanel(weave.Panel):
    id = "SlackMessagesPanel"
    input_node: weave.Node[list[slack.Message]]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.Table:
        messages = typing.cast(list[slack.Message], self.input_node)  # type: ignore
        return weave.legacy.weave.panels.Table(
            messages,
            columns=[
                lambda message: message.user_id,
                lambda message: message.text,
            ],
        )


@weave.type()
class SlackChannelsPanel(weave.Panel):
    id = "SlackChannelsPanel"
    input_node: weave.Node[list[slack.Channel]]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.Table:
        return weave.legacy.weave.panels.Table(
            self.input_node,
            columns=[
                lambda channel: channel.channel_name,
                lambda channel: channel.size(),
            ],
        )


@weave.type()
class SlackChannelPanel(weave.Panel):
    id = "SlackChannelPanel"
    input_node: weave.Node[slack.Channel]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.Card:
        channel = typing.cast(slack.Channel, self.input_node)  # type: ignore
        return weave.legacy.weave.panels.Card(
            title=channel.channel_name,
            subtitle="Slack channel",
            content=[
                weave.legacy.weave.panels.CardTab(
                    name="Messages", content=SlackMessagesPanel(channel.messages())
                ),
            ],
        )


@weave.type()
class SlackPanel(weave.Panel):
    id = "SlackPanel"
    input_node: weave.Node[slack.Slack]

    @weave.op()
    def slack_render(self) -> weave.legacy.weave.panels.Card:
        s = typing.cast(slack.Slack, self.input_node)  # type: ignore
        return weave.legacy.weave.panels.Card(
            title="Slack export data",
            subtitle="",
            content=[
                weave.legacy.weave.panels.CardTab(
                    name="Channels", content=SlackChannelsPanel(s.channels())
                ),
            ],
        )



import logging
import multiprocessing
import queue
import threading
import typing
import warnings

import numpy as np
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

from weave.legacy.weave import errors
from weave.legacy.weave import context_state, environment

umap_lib = {}

DEFAULT_TIMEOUT_SEC = environment.projection_timeout_sec()

ResultQueueItemType = typing.Union[Exception, np.ndarray]


def _get_umap():
    # Lazily load a cached version of UMAP - umap import
    # time is quite expensive so we want to do it once and
    # only when needed
    if "lib" not in umap_lib:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import umap

            umap_lib["lib"] = umap
    return umap_lib["lib"]


def perform_2D_projection(
    np_array_of_embeddings: np.ndarray,
    projectionAlgorithm: str,
    algorithmOptions: dict,
) -> np.ndarray:
    if len(np_array_of_embeddings.shape) != 2:
        raise errors.WeaveInternalError(
            f"The input to the 2D projection must be a 2D array of embeddings, found {np_array_of_embeddings.shape}"
        )
    if projectionAlgorithm == "pca":
        projection = perform_2D_projection_pca(
            np_array_of_embeddings, algorithmOptions.get("pca", {})
        )
    elif projectionAlgorithm == "tsne":
        projection = perform_2D_projection_tsne(
            np_array_of_embeddings, algorithmOptions.get("tsne", {})
        )
    elif projectionAlgorithm == "umap":
        projection = perform_2D_projection_umap(
            np_array_of_embeddings, algorithmOptions.get("umap", {})
        )
    else:
        raise Exception("Unknown projection algorithm: " + projectionAlgorithm)
    return projection


def perform_2D_projection_async(
    np_array_of_embeddings: np.ndarray,
    projectionAlgorithm: str,
    algorithmOptions: dict,
    result_queue: queue.Queue[ResultQueueItemType],
):
    try:
        projection = perform_2D_projection(
            np_array_of_embeddings, projectionAlgorithm, algorithmOptions
        )
        result_queue.put(projection)
    except Exception as e:
        result_queue.put(e)


def perform_2D_projection_with_timeout(
    np_array_of_embeddings: np.ndarray,
    projectionAlgorithm: str,
    algorithmOptions: dict,
    timeout: typing.Optional[typing.Union[int, float]] = DEFAULT_TIMEOUT_SEC,
) -> np.ndarray:
    if timeout is None:
        return perform_2D_projection(
            np_array_of_embeddings, projectionAlgorithm, algorithmOptions
        )

    # otherwise run it in another process and kill it if it goes over time

    result_queue: queue.Queue[ResultQueueItemType] = multiprocessing.Queue()
    target = multiprocessing.Process(
        target=perform_2D_projection_async,
        args=(
            np_array_of_embeddings,
            projectionAlgorithm,
            algorithmOptions,
            result_queue,
        ),
        name="projection",
    )
    target.start()

    try:
        result = result_queue.get(timeout=timeout)
    except queue.Empty:
        target.kill()
        logging.warning(
            f"Projection timed out after {timeout} seconds, killing process {target.pid}, returning empty projection",
        )

        return np.zeros((len(np_array_of_embeddings), 2))
    if isinstance(result, Exception):
        raise result

    return result


def limit_embedding_dimensions(
    np_array_of_embeddings: np.ndarray, max_dimensions: int = 50
) -> np.ndarray:
    max_dimensions = min(max_dimensions, len(np_array_of_embeddings))
    if np_array_of_embeddings.shape[1] > max_dimensions:
        return PCA(n_components=max_dimensions).fit_transform(np_array_of_embeddings)
    return np_array_of_embeddings


def perform_2D_projection_pca(
    np_array_of_embeddings: np.ndarray, options: dict
) -> np.ndarray:
    model = PCA(n_components=2)
    return model.fit_transform(np_array_of_embeddings)


def perform_2D_projection_tsne(
    np_array_of_embeddings: np.ndarray, options
) -> np.ndarray:
    n_samples = len(np_array_of_embeddings)
    return TSNE(
        n_components=2,
        perplexity=min(n_samples - 1, options.get("perplexity", 30)),
        n_iter=min(1000, max(500, options.get("iterations", 1000))),
        learning_rate=options.get("learningRate", "auto"),
        init="random",
    ).fit_transform(limit_embedding_dimensions(np_array_of_embeddings))


# TODO: we need lock stats in datadog
# UMAP is not thread safe, it crashes the server when called at the same
# time on parallel threads.
umap_lock = threading.Lock()


def perform_2D_projection_umap(
    np_array_of_embeddings: np.ndarray, options: dict
) -> np.ndarray:
    model = _get_umap().UMAP(
        n_components=2,
        # Letting the library choose defaults
        # n_neighbors=min(n_samples - 1, options.get("neighbors", 15)),
        # min_dist=options.get("minDist", 0.1),
        # spread=options.get("spread", 1.0),
        init="random",
    )
    with umap_lock:
        return model.fit_transform(np_array_of_embeddings)



from typing import Iterator, Optional

from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
)


class OpenAIStream:
    """
    A class that reconstructs a ChatCompletion from an OpenAI stream.

    Usage: initialize this class, iterate through all the chunks, then
    call final_response to get the final ChatCompletion.

    Args:
        chunk_iter: The output of an openai chat completion streaming=True call.
    """

    def __init__(self, chunk_iter: Iterator[ChatCompletionChunk]) -> None:
        self.chunk_iter = chunk_iter
        self.first_chunk: Optional[ChatCompletionChunk] = None
        self.output_choices: list[dict] = []

    def __next__(self) -> ChatCompletionChunk:
        chunk = self.chunk_iter.__next__()
        self._process_chunk(chunk)
        return chunk

    def __iter__(self) -> Iterator[ChatCompletionChunk]:
        for chunk in self.chunk_iter:
            self._process_chunk(chunk)
            yield chunk

    def _process_chunk(self, chunk: ChatCompletionChunk) -> None:
        if self.first_chunk is None:
            self.first_chunk = chunk
        for chunk_choice in chunk.choices:
            for i in range(chunk_choice.index + 1 - len(self.output_choices)):
                self.output_choices.append(
                    {
                        "index": len(self.output_choices),
                        "message": {
                            "content": None,
                            "tool_calls": None,
                        },
                        "finish_reason": None,
                        "logprobs": None,
                    }
                )

            # choice fields
            choice = self.output_choices[chunk_choice.index]
            if chunk_choice.finish_reason:
                choice["finish_reason"] = chunk_choice.finish_reason
            if chunk_choice.logprobs:
                choice["logprobs"] = chunk_choice.logprobs

            # message
            if chunk_choice.delta.content:
                if choice["message"]["content"] is None:
                    choice["message"]["content"] = ""
                choice["message"]["content"] += chunk_choice.delta.content
            if chunk_choice.delta.role:
                choice["message"]["role"] = chunk_choice.delta.role

            # function call
            if chunk_choice.delta.function_call:
                raise NotImplementedError("Function calls not supported")

            # tool calls
            if chunk_choice.delta.tool_calls:
                if choice["message"]["tool_calls"] is None:
                    choice["message"]["tool_calls"] = []
                for tool_call_delta in chunk_choice.delta.tool_calls:
                    for i in range(
                        tool_call_delta.index + 1 - len(choice["message"]["tool_calls"])  # type: ignore
                    ):
                        choice["message"]["tool_calls"].append(  # type: ignore
                            {
                                "function": {"name": None, "arguments": ""},
                            }
                        )
                    tool_call = choice["message"]["tool_calls"][  # type: ignore
                        tool_call_delta.index
                    ]
                    if tool_call_delta.id is not None:
                        tool_call["id"] = tool_call_delta.id
                    if tool_call_delta.type is not None:
                        tool_call["type"] = tool_call_delta.type
                    if tool_call_delta.function is not None:
                        if tool_call_delta.function.name is not None:
                            tool_call["function"]["name"] = (
                                tool_call_delta.function.name
                            )
                        if tool_call_delta.function.arguments is not None:
                            tool_call["function"]["arguments"] += (
                                tool_call_delta.function.arguments
                            )

    def final_response(self) -> ChatCompletion:
        if self.first_chunk is None:
            raise ValueError("No chunks received")
        return ChatCompletion(
            id=self.first_chunk.id,
            choices=self.output_choices,  # type: ignore
            created=self.first_chunk.created,
            model=self.first_chunk.model,
            object="chat.completion",
        )



import typing

import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave.panels_py.generator_templates import template_registry


@weave.type()
class PyBoardSeedBoardConfig:
    pass


@weave.op(  # type: ignore
    name="py_board-seed_board",
    hidden=True,
)
def seed_board(
    input_node: weave.Node[typing.Any],
    config: typing.Optional[PyBoardSeedBoardConfig] = None,
) -> weave.legacy.weave.panels.Group:
    control_items = [
        weave.legacy.weave.panels.GroupPanel(
            input_node,
            id="data",
        ),
    ]

    panels = [
        weave.legacy.weave.panels.BoardPanel(
            weave_internal.make_var_node(input_node.type, "data"),
            id="panel",
            layout=weave.legacy.weave.panels.BoardPanelLayout(x=0, y=0, w=24, h=6),
        ),
    ]
    return weave.legacy.weave.panels.Board(vars=control_items, panels=panels)


template_registry.register(
    "py_board-seed_board",
    "Simple Board",
    "Seed a board with a simple visualization of this table.",
)



import asyncio
import dataclasses
import datetime
import json
import logging
import typing

from weave.legacy.weave import weave_internal
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op, weave_class
from weave.legacy.weave import (
    artifact_fs,
    artifact_wandb,
    io_service,
    ops_arrow,
    wandb_util,
    engine_trace,
    errors, 
    util, 
)
from weave.legacy.weave import timestamp as weave_timestamp
from weave.legacy.weave.ops_domain import trace_tree, wbmedia


@dataclasses.dataclass(frozen=True)
class TableType(types.ObjectType):
    name = "table"

    def property_types(self):
        return {
            "_rows": ops_arrow.ArrowWeaveListType(types.TypedDict({})),
        }


@weave_class(weave_type=TableType)
class Table:
    def __init__(self, _rows):
        self._rows = _rows

    @op(
        name="table-rowsType",
        input_type={"table": types.optional(TableType())},
        output_type=types.TypeType(),
        hidden=True,
    )
    def rows_type(table):
        if table == None:
            return types.NoneType()
        ttype = types.TypeRegistry.type_of(table._rows)
        return ttype

    @op(
        name="table-rows",
        input_type={"table": TableType()},
        output_type=ops_arrow.ArrowWeaveListType(types.TypedDict({})),
        refine_output_type=rows_type,
    )
    def rows(table):
        return table._rows


@dataclasses.dataclass(frozen=True)
class PartitionedTableType(types.ObjectType):
    name = "partitioned-table"

    def property_types(self):
        return {
            "_rows": ops_arrow.ArrowWeaveListType(types.TypedDict({})),
            "_file": artifact_fs.FilesystemArtifactFileType(),
            "partitionedTable": types.TypedDict({"parts_path": types.String()}),
        }


@weave_class(weave_type=PartitionedTableType)
class PartitionedTable:
    def __init__(self, _rows, _file, partitionedTable):
        self._rows = _rows
        self._file = _file
        self.partitionedTable = partitionedTable

    @op(
        name="partitionedtable-file",
        input_type={"partitionedTable": PartitionedTableType()},
        output_type=artifact_fs.FilesystemArtifactFileType(),
    )
    def partitioned_file(partitionedTable):
        return partitionedTable._file

    @op(
        name="partitionedtable-rowsType",
        input_type={"partitionedtable": PartitionedTableType()},
        output_type=types.TypeType(),
        hidden=True,
    )
    def partitioned_rows_type(partitionedtable):
        ttype = types.TypeRegistry.type_of(partitionedtable._rows)
        return ttype

    @op(
        name="partitionedtable-rows",
        input_type={"partitionedtable": PartitionedTableType()},
        output_type=ops_arrow.ArrowWeaveListType(types.TypedDict({})),
        refine_output_type=partitioned_rows_type,
    )
    def rows(partitionedtable):
        return partitionedtable._rows


@dataclasses.dataclass(frozen=True)
class JoinedTableType(types.ObjectType):
    name = "joined-table"

    def property_types(self):
        return {
            "_rows": ops_arrow.ArrowWeaveListType(types.TypedDict({})),
            "_file": artifact_fs.FilesystemArtifactFileType(),
        }


@weave_class(weave_type=JoinedTableType)
class JoinedTable:
    def __init__(self, _rows, _file):
        self._rows = _rows
        self._file = _file

    @op(
        name="joinedtable-file",
        input_type={"joinedTable": JoinedTableType()},
        output_type=artifact_fs.FilesystemArtifactFileType(),
    )
    def joined_file(joinedTable):
        return joinedTable._file

    @op(
        name="joinedtable-rowsType",
        input_type={"joinedtable": JoinedTableType()},
        output_type=types.TypeType(),
        hidden=True,
    )
    def joined_rows_type(joinedtable):
        ttype = types.TypeRegistry.type_of(joinedtable._rows)
        return ttype

    @op(
        name="joinedtable-rows",
        input_type={"joinedtable": JoinedTableType()},
        output_type=ops_arrow.ArrowWeaveListType(types.TypedDict({})),
        refine_output_type=joined_rows_type,
    )
    def rows(joinedtable):
        return joinedtable._rows


def _data_is_legacy_run_file_format(data):
    return "columns" in data and "data" in data


def _data_is_weave_file_format(data):
    return "columns" in data and "data" in data and "column_types" in data


def _data_is_weave_file_with_mixed_type_settings(data):
    # This is a special case that occurs when the user constructs a Table with
    # "allow_mixed_types=True". In these cases the client does not report the
    # type of the data, which results in all columns having the type
    # Optional<Any>. When this occurs, we have absolutely no type information.
    if not _data_is_weave_file_format(data):
        return False
    column_types = data["column_types"]
    type_map = column_types.get("params", {}).get("type_map")
    if not type_map:
        return False
    return all(
        v
        == {
            "params": {"allowed_types": [{"wb_type": "any"}, {"wb_type": "none"}]},
            "wb_type": "union",
        }
        for v in type_map.values()
    )


def _infer_type_from_cell(cell: typing.Any) -> types.Type:
    if isinstance(cell, dict) and "_type" in cell and isinstance(cell["_type"], str):
        maybe_type = types.type_name_to_type(cell["_type"])
        if maybe_type is not None:
            return maybe_type()
    return types.TypeRegistry.type_of(cell)


def _infer_type_from_row_dict(row: dict) -> types.Type:
    return types.TypedDict({k: _infer_type_from_cell(v) for k, v in row.items()})


def _infer_type_from_col_list(row: list) -> types.Type:
    if len(row) == 0:
        return types.NoneType()
    running_type: types.Type = types.UnknownType()
    for cell in row:
        running_type = types.merge_types(running_type, _infer_type_from_cell(cell))
    return running_type


def _infer_type_from_row_dicts(rows: list[dict]) -> types.TypedDict:
    if len(rows) == 0:
        return types.TypedDict({})
    running_type: types.Type = types.UnknownType()
    for row in rows:
        running_type = types.merge_types(running_type, _infer_type_from_row_dict(row))
    if not isinstance(running_type, types.TypedDict):
        raise errors.WeaveInternalError(
            f"Expected running_type to be a TypedDict, but got {running_type}"
        )
    return running_type


@dataclasses.dataclass
class PeerTableReader:
    """
    Provides a simple interface for reading rows from a peer table. Peer tables
    can be read via index - which is straight forward. Peer tables can be read
    via a column key - which we want to return the first row which matches this
    value (think of it like a join but just getting the first value.) We can
    optimize this by saving the reverse index map while reading and only read
    the necessary rows. This means we will only visit each "cell" in the table
    at most 1 time.
    """

    peer_rows: list[dict]
    peer_object_type: types.TypedDict
    lookup_map: dict[str, dict[str, int]] = dataclasses.field(default_factory=dict)
    scan_loc: dict[str, int] = dataclasses.field(default_factory=dict)

    def row_at_index(self, index: int) -> dict:
        return self.peer_rows[index]

    def row_at_key(self, column: str, key: str) -> typing.Optional[dict]:
        # First, we create a default index map for this column if it doesn't
        # exist.
        if column not in self.lookup_map:
            self.lookup_map[column] = {}
            self.scan_loc[column] = 0
        index_map = self.lookup_map[column]

        # If we have already seen this key, we can just return the row
        if key in index_map:
            return self.peer_rows[index_map[key]]

        # Otherwise, we scan the table until we find the key
        # creating a reverse index map as we go. This allows
        # us to break early, and continue the scan later if needed.
        while self.scan_loc[column] < len(self.peer_rows):
            curr_row = self.peer_rows[self.scan_loc[column]]
            cell_val = curr_row[column]
            if cell_val not in index_map:
                index_map[cell_val] = self.scan_loc[column]
            if cell_val == key:
                return curr_row
            self.scan_loc[column] += 1
        return None


def _in_place_join_in_linked_tables(
    rows: list[dict],
    object_type: types.TypedDict,
    column_types: wandb_util.Weave0TypeJson,
    file: artifact_fs.FilesystemArtifactFile,
) -> typing.Tuple[list[dict], types.TypedDict]:
    """
    This function will join in any peer linked tables. This is done by replacing
    the column with the source table's index (or key) with the peer table's row.
    This is done in place, and the object type is updated to reflect the new
    column types.

    Note: We are currently doing this "eagerly". However, it is reasonable to
    assume that once we can read table types quickly, we can differ this
    calculate in the form of a ref.
    """
    type_map = column_types["params"]["type_map"]
    for column_name, column_type in type_map.items():
        col_wb_type = column_type["wb_type"]
        if (
            col_wb_type in wandb_util.foreign_key_type_names
            or col_wb_type in wandb_util.foreign_index_type_names
        ):
            peer_file = file.artifact.path_info(column_type["params"]["table"])
            if peer_file is None:
                # if is none, then the file doesn't exist, so we can't join.
                # We will just set the column to an empty dict. This is preferred
                # over failing.
                peer_object_type = types.TypedDict({})
                # Update the row values
                for row in rows:
                    row[column_name] = {}
            elif isinstance(peer_file, artifact_fs.FilesystemArtifactDir):
                raise errors.WeaveInternalError("Peer file is a directory")
            else:
                with peer_file.open() as f:
                    tracer = engine_trace.tracer()
                    with tracer.trace("peer_table:jsonload"):
                        peer_data = json.load(f)
                (
                    peer_rows,
                    peer_object_type,
                ) = _get_rows_and_object_type_from_weave_format(peer_data, peer_file)
                peer_reader = PeerTableReader(peer_rows, peer_object_type)

                # Update the row values
                for row in rows:
                    if col_wb_type in wandb_util.foreign_index_type_names:
                        row[column_name] = peer_reader.row_at_index(row[column_name])
                    else:
                        row[column_name] = peer_reader.row_at_key(
                            column_type["params"]["col_name"], row[column_name]
                        )

            # update the object type
            object_type.property_types[column_name] = peer_object_type

    return rows, object_type


def _make_type_non_none(t: types.Type) -> types.Type:
    non_none_type = types.non_none(t)
    if isinstance(non_none_type, types.List):
        return types.List(_make_type_non_none(non_none_type.object_type))
    elif isinstance(non_none_type, types.TypedDict):
        return types.TypedDict(
            {k: _make_type_non_none(v) for k, v in non_none_type.property_types.items()}
        )
    return non_none_type


possible_media_type_classes = (
    wbmedia.LegacyImageArtifactFileRefType,
    wbmedia.ImageArtifactFileRefType,
    wbmedia.AudioArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.BokehArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.VideoArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.Object3DArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.MoleculeArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.HtmlArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.LegacyTableNDArrayType,
)


def _table_data_to_weave1_objects(
    row_data: list[dict],
    file: artifact_fs.FilesystemArtifactFile,
    row_type: types.TypedDict,
) -> list[dict]:
    def _create_media_type_for_cell(cell: dict) -> typing.Any:
        file_type = cell["_type"]
        file_path = cell["path"]
        if file_type == "image-file":
            return wbmedia.ImageArtifactFileRef(
                artifact=file.artifact,
                path=file_path,
                format=cell["format"],
                height=cell.get("height", 0),
                width=cell.get("width", 0),
                sha256=cell.get("sha256", file_path),
                boxes=cell.get("boxes", {}),  # type: ignore
                masks=cell.get("masks", {}),  # type: ignore
                classes=cell.get("classes"),  # type: ignore
            )
        elif file_type in [
            "audio-file",
            "bokeh-file",
            "video-file",
            "object3D-file",
            "molecule-file",
            "html-file",
        ]:
            type_cls = types.type_name_to_type(file_type)
            if type_cls is not None and type_cls.instance_class is not None:
                return type_cls.instance_class(
                    file.artifact, file_path, cell.get("sha256", file_path)
                )
        else:
            raise errors.WeaveTableDeserializationError(
                f"Unsupported media type {file_type}"
            )

    def _process_cell_value(cell: typing.Any, cell_type: types.Type) -> typing.Any:
        if (
            isinstance(cell, list)
            and isinstance(cell_type, types.List)
            # We want to avoid recursing into the list if the list element type is basic,
            # since it will be a relatively expensive `O(n)` no-op.
            and not isinstance(cell_type.object_type, types.BasicType)
        ):
            cell = [_process_cell_value(c, cell_type.object_type) for c in cell]

        elif isinstance(cell, dict) and isinstance(cell_type, types.TypedDict):
            cell = {
                k: _process_cell_value(v, cell_type.property_types[str(k)])
                for k, v in cell.items()
            }
        # this is needed because tables store timestamps as unix epochs in ms, but the deserialized
        # representation in weave1 is a datetime object. we do the conversion here to ensure deserialized
        # timestamps all expose a common interface to weave1 callers.
        elif isinstance(cell_type, types.Timestamp):
            if cell is not None:
                cell = weave_timestamp.ms_to_python_datetime(
                    weave_timestamp.unitless_int_to_inferred_ms(cell)
                )
        elif isinstance(cell, dict):
            if isinstance(cell_type, possible_media_type_classes):
                cell = _create_media_type_for_cell(cell)
            elif isinstance(cell_type, trace_tree.WBTraceTree.WeaveType):  # type: ignore
                copy = {**cell}
                copy.pop("_type")
                cell = trace_tree.WBTraceTree(**copy)

        return cell

    non_null_prop_types = _make_type_non_none(row_type)
    non_null_prop_types = typing.cast(types.TypedDict, non_null_prop_types)

    rows: list[dict] = []
    for data_row in row_data:
        new_row = {
            k: _process_cell_value(v, non_null_prop_types.property_types[str(k)])
            for k, v in data_row.items()
        }
        rows.append(new_row)

    return rows


def _patch_legacy_image_file_types(
    rows: list[dict],
    object_type: types.TypedDict,
    file: artifact_fs.FilesystemArtifactFile,
    assume_legacy: bool = False,
):
    """
    This is a patch to make up for image files born before Oct 2021. Before then, we did not encode the
    full box and mask type information in the type system and unfortunately requires looking at the data.
    We will only look at the first 10 examples, however. This is a rewrite of the same patch function found here:
    https://github.com/wandb/core/blob/30d6d9eeb125904c00066f023ac404c957bf6bf7/lib/js/cg/src/ops/domain/table.ts#LL51C29-L51C29

    Note: this does not work for lists of images!
    """
    return_prop_types = {}
    class_types = (
        (wbmedia.LegacyImageArtifactFileRefType,)
        if not assume_legacy
        else (wbmedia.LegacyImageArtifactFileRefType, wbmedia.ImageArtifactFileRefType)
    )
    for col_name, col_type in object_type.property_types.items():
        return_prop_types[col_name] = col_type
        other_members = []

        # If the type is a union, we need to check if the union contains a legacy image type
        # If it does, we keep track of the peer types so we can re-create the union
        if isinstance(col_type, types.UnionType):
            found = False
            for member in col_type.members:
                if not isinstance(member, class_types):
                    other_members.append(member)
                else:
                    found = True
            if not found:
                continue

        # If the type is not a union, we need to check if it is a legacy image type
        elif not isinstance(col_type, class_types):
            continue

        # We will look at the first 10 rows of the table to determine the type.
        non_null_rows_evaluated = 0
        mask_keys_set: dict[str, bool] = {}
        box_keys_set: dict[str, bool] = {}
        box_score_key_set: dict[str, bool] = {}
        class_map: dict[int, str] = {}
        for row in rows:
            image_example = row[col_name]
            if image_example is None:
                continue
            image_example = typing.cast(wbmedia.ImageArtifactFileRef, image_example)

            # Add the mask keys to the running set
            if image_example.masks is not None:
                mask_keys_set.update(dict.fromkeys(image_example.masks.keys()))

            # Add the box keys to the running set
            if image_example.boxes is not None:
                box_keys_set.update(dict.fromkeys(image_example.boxes.keys()))

                # Add the box score keys to the running set
                for box_set in image_example.boxes.values():
                    for box in box_set:
                        if "scores" in box and box["scores"] is not None:
                            box_score_key_set.update(
                                dict.fromkeys(box["scores"].keys())
                            )

            # Fetch the class data (requires a network call) and update the class map
            if image_example.classes is not None and "path" in image_example.classes:
                classes_path = image_example.classes["path"]
                classes_file = file.artifact.path_info(classes_path)
                if classes_file is None or isinstance(
                    classes_file, artifact_fs.FilesystemArtifactDir
                ):
                    raise errors.WeaveInternalError(
                        "classes_file is None or a directory"
                    )
                with classes_file.open() as f:
                    tracer = engine_trace.tracer()
                    with tracer.trace("classes_file:jsonload"):
                        classes_data = json.load(f)
                if (
                    "class_set" in classes_data
                    and classes_data["class_set"] is not None
                ):
                    for class_item in classes_data["class_set"]:
                        if (
                            "id" in class_item
                            and "name" in class_item
                            and class_map.get(class_item["id"]) is None
                        ):
                            class_map[class_item["id"]] = class_item["name"]

            # Update the number of loaded rows
            non_null_rows_evaluated += 1
            if non_null_rows_evaluated >= 10:
                break

        # Finally, construct the needed objects
        class_keys = list(class_map.keys())
        box_layers = {box_key: class_keys for box_key in box_keys_set}
        mask_layers = {box_key: class_keys for box_key in mask_keys_set}

        new_image_type = wbmedia.ImageArtifactFileRefType(
            boxLayers=box_layers,
            boxScoreKeys=list(box_score_key_set),
            maskLayers=mask_layers,
            classMap=class_map,
        )

        return_prop_types[col_name] = types.union(*other_members, new_image_type)
    return types.TypedDict(return_prop_types)


def should_infer_type_from_data(col_type: types.Type) -> bool:
    status = {"found_unknown": False}

    def is_unknown_type_mapper(t: types.Type) -> None:
        status["found_unknown"] = status["found_unknown"] or isinstance(
            t, (types.UnknownType, types.Any)
        )

    types.map_leaf_types(col_type, is_unknown_type_mapper)

    return status["found_unknown"]


def _get_rows_and_object_type_from_weave_format(
    data: typing.Any,
    file: artifact_fs.FilesystemArtifactFile,
    sample_max_rows: int = 1000,
) -> tuple[list, types.TypedDict]:
    rows = []
    artifact = file.artifact
    if not isinstance(artifact, artifact_wandb.WandbArtifact):
        raise errors.WeaveInternalError(
            "Weave table file format is only supported for wandb artifacts"
        )
    row_data = data["data"]
    # Always convert the column names to string names
    column_types = data["column_types"]
    column_names = [str(c) for c in data["columns"]]
    converted_object_type = wandb_util.weave0_type_json_to_weave1_type(column_types)
    if not isinstance(converted_object_type, types.TypedDict):
        raise errors.WeaveInternalError(
            "Weave table file format only supports typed dicts"
        )
    converted_object_type = types.TypedDict(
        {str(k): v for k, v in converted_object_type.property_types.items()}
    )
    # Fix two things:
    # 1. incoming table column names may not match the order of column_types
    # 2. if we had an unknown (happens when old type is "PythonObjectType")
    #    we need to manually detect the type.
    obj_prop_types = {}
    for i, key in enumerate(column_names):
        if key not in converted_object_type.property_types:
            raise errors.WeaveTableDeserializationError(
                f"Column name {key} not found in column_types"
            )
        col_type = converted_object_type.property_types[key]
        if should_infer_type_from_data(col_type):
            # Sample some data to detect the type. Otherwise this
            # can be very expensive. This could cause down-stream crashes,
            # for example if we don't realize that a column is union of string
            # and int, saving to arrow will crash.
            unknown_col_example_data = [
                row[i] for row in util.sample_rows(row_data, sample_max_rows)
            ]
            obj_prop_types[key] = _infer_type_from_col_list(unknown_col_example_data)
            logging.warning(
                f"Column {key} had type {col_type} requiring data-inferred type. Inferred type as {obj_prop_types[key]}. This may be incorrect due to data sampling"
            )
        else:
            obj_prop_types[key] = col_type
    object_type = types.TypedDict(obj_prop_types)

    raw_rows = [dict(zip(column_names, row)) for row in row_data]
    rows = _table_data_to_weave1_objects(raw_rows, file, object_type)
    object_type = _patch_legacy_image_file_types(rows, object_type, file)

    rows, object_type = _in_place_join_in_linked_tables(
        rows, object_type, column_types, file
    )

    return rows, object_type


def _get_rows_and_object_type_from_weave_format_mixed(
    data: typing.Any,
    file: artifact_fs.FilesystemArtifactFile,
    sample_max_rows: int = 1000,
) -> tuple[list, types.TypedDict]:
    # In this case (see notes on
    # `_data_is_weave_file_with_mixed_type_settings`), we have no useful type
    # information available. If the data is small, we could dispatch to the
    # `_get_rows_and_object_type_from_legacy_format`, however, in practice this
    # can be prohibitively expensive. Instead, we will just convert all of the
    # columns to strings and give a slightly better treatment in the UI. This is
    # certainly a sub optimal situation because there could be cases where users
    # set this setting to true, yet their types are homogenous enough where the
    # old logic would be able to parse the content and make sense of it.
    # However, in the degenerate cases, this is prohibitively expensive to
    # parse, could error out, and the data is not even useful for the user.
    # Let's at least give them a string representation of the data and ensure
    # that it will be displayed.
    rows = [
        {
            col_key: json.dumps(r[col_ndx])
            if type(r[col_ndx]) in [list, dict]
            else r[col_ndx]
            for col_ndx, col_key in enumerate(data["columns"])
        }
        for r in data["data"]
    ]
    sampled_rows = util.sample_rows(rows, sample_max_rows)
    awl = ops_arrow.to_arrow(sampled_rows, None, file.artifact)
    return rows, awl.object_type


def _get_rows_and_object_type_from_legacy_format(
    data: dict, file: artifact_fs.FilesystemArtifactFile, sample_max_rows: int = 1000
) -> tuple[list, types.TypedDict]:
    # W&B dataframe columns are ints, we always want strings
    data["columns"] = [str(c) for c in data["columns"]]
    raw_rows = [dict(zip(data["columns"], row)) for row in data["data"]]
    object_type = _infer_type_from_row_dicts(
        util.sample_rows(raw_rows, sample_max_rows)
    )

    rows = _table_data_to_weave1_objects(raw_rows, file, object_type)
    object_type = _patch_legacy_image_file_types(rows, object_type, file, True)

    return rows, object_type


@dataclasses.dataclass
class _TableLikeAWLFromFileResult:
    awl: ops_arrow.ArrowWeaveList
    data: dict


def _get_table_data_from_file(file: artifact_fs.FilesystemArtifactFile) -> dict:
    tracer = engine_trace.tracer()
    if file is None or isinstance(file, artifact_fs.FilesystemArtifactDir):
        raise errors.WeaveInternalError("File is None or a directory")
    with file.open() as f:
        with tracer.trace("get_table:jsonload"):
            data = json.load(f)
    return data


def _get_table_like_awl_from_file(
    file: typing.Union[
        artifact_fs.FilesystemArtifactFile, artifact_fs.FilesystemArtifactDir, None
    ],
    num_parts: int = 1,
) -> _TableLikeAWLFromFileResult:
    if file is None or isinstance(file, artifact_fs.FilesystemArtifactDir):
        raise errors.WeaveInternalError("File is None or a directory")
    data = _get_table_data_from_file(file)
    if file.path.endswith(".joined-table.json"):
        awl = _get_joined_table_awl_from_file(data, file)
    elif file.path.endswith(".partitioned-table.json"):
        awl = _get_partitioned_table_awl_from_file(data, file)
    elif file.path.endswith(".table.json"):
        awl = _get_table_awl_from_file(data, file, num_parts)
    else:
        raise errors.WeaveInternalError(
            f"Unknown table file format for path: {file.path}"
        )
    return _TableLikeAWLFromFileResult(awl, data)


USE_RAW_DATA_WHEN_MIXED = False


def _get_rows_and_object_type_awl_from_file(
    data: dict,
    file: artifact_fs.FilesystemArtifactFile,
    num_parts: int = 1,
) -> typing.Tuple[list, types.Type]:
    tracer = engine_trace.tracer()
    rows: list = []
    object_type = None
    with tracer.trace("get_table:get_rows_and_object_type"):
        sample_max_rows = max(1000 // num_parts, 1)
        if USE_RAW_DATA_WHEN_MIXED and _data_is_weave_file_with_mixed_type_settings(
            data
        ):
            # `_data_is_weave_file_format` is known to be fallible, especially
            # with large datasets of heterogenous types. In this case, the user
            # explicitly told us that the types are mixed, so we will just
            # convert everything to strings.
            rows, object_type = _get_rows_and_object_type_from_weave_format_mixed(
                data, file, sample_max_rows
            )
        elif _data_is_weave_file_format(data):
            rows, object_type = _get_rows_and_object_type_from_weave_format(
                data, file, sample_max_rows
            )
        elif _data_is_legacy_run_file_format(data):
            rows, object_type = _get_rows_and_object_type_from_legacy_format(
                data, file, sample_max_rows
            )
        else:
            raise errors.WeaveInternalError("Unknown table file format for data")

    return rows, object_type


def _get_table_awl_from_rows_object_type(
    rows: list, object_type: types.Type, file: artifact_fs.FilesystemArtifactFile
) -> "ops_arrow.ArrowWeaveList":
    tracer = engine_trace.tracer()
    with tracer.trace("get_table:to_arrow"):
        return ops_arrow.to_arrow_from_list_and_artifact(
            rows, object_type, file.artifact
        )


def _get_table_awl_from_file(
    data: dict, file: artifact_fs.FilesystemArtifactFile, num_parts: int = 1
) -> "ops_arrow.ArrowWeaveList":
    rows, object_type = _get_rows_and_object_type_awl_from_file(data, file, num_parts)
    return _get_table_awl_from_rows_object_type(rows, object_type, file)


def _get_partitioned_table_awl_from_file(
    data: dict, file: artifact_fs.FilesystemArtifactFile
) -> ops_arrow.ArrowWeaveList:
    parts_path_root = data["parts_path"]

    all_aws: list[ops_arrow.ArrowWeaveList] = []
    part_dir = file.artifact.path_info(parts_path_root)
    if isinstance(part_dir, artifact_fs.FilesystemArtifactDir):
        # Pre-download all the files in parallel.
        # We do this because we currently only have a synchronous pattern
        # available for resolving artifact-backed files.
        # TODO: Remove pre-download once artifact-backed files can be resolved asynchronously
        asyncio.run(ensure_files(part_dir.files))

        num_parts = len(part_dir.files)
        rrows: list[list] = []
        object_types: list[types.Type] = []
        for file in part_dir.files.values():
            data = _get_table_data_from_file(file)
            rows, object_type = _get_rows_and_object_type_awl_from_file(
                data, file, num_parts
            )
            rrows.append(rows)
            object_types.append(object_type)
        object_type = types.union(*object_types)

        for rows, file in zip(rrows, part_dir.files.values()):
            all_aws.append(
                _get_table_awl_from_rows_object_type(rows, object_type, file)
            )
    arrow_weave_list = ops_arrow.ops.concat.raw_resolve_fn(all_aws)
    return arrow_weave_list


# Download files in a `FilesystemArtifactDir` in parallel.
# This only downloads files that are `WandbArtifact`s and have a resolved `_read_artifact_uri`.
async def ensure_files(files: dict[str, artifact_fs.FilesystemArtifactFile]):
    client = io_service.get_async_client()

    loop = asyncio.get_running_loop()

    tasks = set()
    async with client.connect() as conn:
        for file in files.values():
            if (
                isinstance(file.artifact, artifact_wandb.WandbArtifact)
                and file.artifact._read_artifact_uri
            ):
                uri = file.artifact._read_artifact_uri.with_path(file.path)
                task = loop.create_task(conn.ensure_file(uri))
                tasks.add(task)
        await asyncio.wait(tasks)


def _get_joined_table_awl_from_file(
    data: dict, file: artifact_fs.FilesystemArtifactFile
) -> ops_arrow.ArrowWeaveList:
    join_key = data["join_key"]

    table_1_path = data["table1"]
    table_2_path = data["table2"]

    awl_1 = _get_table_like_awl_from_file(file.artifact.path_info(table_1_path)).awl
    awl_2 = _get_table_like_awl_from_file(file.artifact.path_info(table_2_path)).awl

    join_fn_1 = weave_internal.define_fn(
        {"row": awl_1.object_type}, lambda row: row[join_key]
    )
    join_fn_2 = weave_internal.define_fn(
        {"row": awl_2.object_type}, lambda row: row[join_key]
    )

    # Note: in WeaveJS, we allow the user to specify the join type, but
    # in practice it is always a full-outer join. If we want to parameterize that
    # then we need to filter out the unneeded rows in joinedTable-rows since we
    # eagerly construct the rows here.
    arrow_weave_list = ops_arrow.list_join.join_2.raw_resolve_fn(
        awl_1, awl_2, join_fn_1.val, join_fn_2.val, "0", "1", True, True
    )
    return arrow_weave_list


@op(name="file-table")
def file_table(file: artifact_fs.FilesystemArtifactFile) -> typing.Optional[Table]:
    # We do a try/catch here because it is possible that the `file` does not
    # exist. This can happen if the user has deleted the referenced artifact. In
    # this case we will throw a `FileNotFoundError` when calling file.open().
    # Technically we probably should do this try/catch in every case of
    # `file.open` (for example in file-contents, file-media, etc...). However,
    # in practice, the `file.open` should only fail if the File is constructed
    # directly (as opposed to use the `artifact.path_info` method - since
    # `path_info` will return None in this case). This only happens when pulling
    # a file from run summary since it is constructed from a hard-coded artifact
    # URL. This path is only used when fetching tables.
    try:
        return Table(_get_table_like_awl_from_file(file).awl)
    except FileNotFoundError as e:
        return None


@op(name="file-partitionedTable")
def partitioned_table(
    file: artifact_fs.FilesystemArtifactFile,
) -> typing.Optional[PartitionedTable]:
    # Please see comment in `file_table` regarding try/catch
    try:
        res = _get_table_like_awl_from_file(file)
    except FileNotFoundError as e:
        return None
    return PartitionedTable(res.awl, file, res.data)


@op(name="file-joinedTable")
def joined_table(
    file: artifact_fs.FilesystemArtifactFile,
) -> typing.Optional[JoinedTable]:
    # Please see comment in `file_table` regarding try/catch
    try:
        return JoinedTable(_get_table_like_awl_from_file(file).awl, file)
    except FileNotFoundError as e:
        return None



import datetime
import json

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op
from weave.legacy.weave.gql_json_cache import use_json
from weave.legacy.weave.gql_op_plugin import wb_gql_op_plugin
from weave.legacy.weave.ops_domain.wandb_domain_gql import (
    _make_alias,
)

rpt_op_configs = {
    "weekly_users_by_country_by_repo": types.TypedDict(
        {
            "user_fraction": types.Number(),
            "country": types.String(),
            "created_week": types.Timestamp(),
            "framework": types.String(),
        }
    ),
    "weekly_repo_users_by_persona": types.TypedDict(
        {
            "created_week": types.Timestamp(),
            "framework": types.String(),
            "persona": types.String(),
            "percentage": types.Number(),
        }
    ),
    "weekly_engaged_user_count_by_repo": types.TypedDict(
        {
            "created_week": types.Timestamp(),
            "framework": types.String(),
            "user_count": types.Number(),
        }
    ),
    "repo_gpu_backends": types.TypedDict(
        {
            "created_week": types.Timestamp(),
            "framework": types.String(),
            "gpu": types.String(),
            "percentage": types.Number(),
        }
    ),
    "versus_other_repos": types.TypedDict(
        {
            "created_week": types.Timestamp(),
            "framework": types.String(),
            "percentage": types.Number(),
        }
    ),
    "runtime_buckets": types.TypedDict(
        {
            "created_week": types.Timestamp(),
            "framework": types.String(),
            "bucket": types.String(),
            "bucket_run_percentage": types.Number(),
        }
    ),
    "user_model_train_freq": types.TypedDict(
        {
            "created_week": types.Timestamp(),
            "framework": types.String(),
            "train_freq": types.String(),
            "percentage": types.Number(),
        }
    ),
    "runs_versus_other_repos": types.TypedDict(
        {
            "created_week": types.Timestamp(),
            "framework": types.String(),
            "percentage": types.Number(),
        }
    ),
    "product_usage": types.TypedDict(
        {
            "created_week": types.Timestamp(),
            "framework": types.String(),
            "product": types.String(),
            "percentage": types.Number(),
        }
    ),
}


def make_rpt_op(plot_name, output_row_type):
    output_type = types.TypedDict(
        {
            "rows": types.List(output_row_type),
            "isNormalizedUserCount": types.Boolean(),
        }
    )

    @op(
        name=f"rpt_{plot_name}GQLResolver",
        input_type={"gql_result": types.TypedDict({}), "repoName": types.String()},
        output_type=output_type,
        hidden=True,
    )
    def root_all_projects_gql_resolver(gql_result, repoName):
        # Copied from root.ts
        alias = _make_alias(
            repoName,
            plot_name,
            "first: 100000",
            prefix="repoInsightsPlotData",
        )
        res = gql_result[alias]
        raw_rows = [use_json(edge["node"]["row"]) for edge in res["edges"]]
        schema_str = res.get("schema", "[]")
        schema = use_json(schema_str)

        is_normalized_user_count = res.get("isNormalizedUserCount", False)

        if not schema:
            raise errors.WeaveInternalError(f"No schema for {alias}")

        def process_row(row):
            processed_row = {}
            for i in range(len(schema)):
                name = schema[i]["Name"]
                type = schema[i]["Type"]
                if type == "TIMESTAMP":
                    processed_row[name] = datetime.datetime.fromtimestamp(row[i])
                else:
                    processed_row[name] = row[i]
            return processed_row

        rows = [process_row(row) for row in raw_rows]

        return {
            "rows": rows,
            "isNormalizedUserCount": is_normalized_user_count,
        }

    def plugin_fn(inputs, inner):
        alias = _make_alias(
            inputs.raw["repoName"],
            plot_name,
            "first: 100000",
            prefix="repoInsightsPlotData",
        )
        return f"""
            {alias}: repoInsightsPlotData(plotName: {json.dumps(plot_name)}, repoName: {inputs["repoName"]}, first: 100000) {{
                edges {{
                    node {{
                        row
                    }}
                }}
                schema
                isNormalizedUserCount
            }}"""

    @op(
        name=f"rpt_{plot_name}",
        input_type={"repoName": types.String()},
        output_type=output_type,
        plugins=wb_gql_op_plugin(
            plugin_fn,
            is_root=True,
            root_resolver=root_all_projects_gql_resolver,
        ),
        hidden=True,
    )
    def root_rpt(repoName):
        raise errors.WeaveGQLCompileError(
            "root-allProjects should not be executed directly. If you see this error, it is a bug in the Weave compiler."
        )

    return root_rpt


# Make all the ops!
for plot_name, output_row_type in rpt_op_configs.items():
    make_rpt_op(plot_name, output_row_type)


@op(
    name="normalizeUserCounts",
    input_type={
        "arr": types.List(
            types.TypedDict(
                {
                    "created_week": types.Timestamp(),
                    "user_count": types.Number(),
                }
            )
        ),
        "normalize": types.Boolean(),
    },
    output_type=lambda input_types: input_types["arr"],
)
def normalize_user_counts(arr, normalize):
    if len(arr) == 0 or not normalize:
        return arr
    min_date = min(arr, key=lambda x: x["created_week"])["created_week"]
    norm_factor = sum(x["user_count"] for x in arr if x["created_week"] == min_date)
    res = []
    for row in arr:
        r = row.copy()
        r["user_count"] = row["user_count"] / norm_factor
        res.append(r)
    return res



import itertools
import textwrap
import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import forward_graph, graph


class _CombinedConstVal:
    """Special value to respresent the combination of multiple values in debug output."""

    def __init__(self, vals: typing.Sequence[typing.Any]):
        self.vals = vals

    def __str__(self) -> str:
        s_vals = [v.__repr__() for v in self.vals]
        s_vals_uniq = sorted(dict.fromkeys(s_vals).keys())
        return "EACH[%s]" % ", ".join(s_vals_uniq)


def combine_common_nodes(outputs: list[graph.Node]) -> list[graph.Node]:
    """Simplifies a dag or set of dags by combining similar structures.

    This produces Nodes, but they are no longer executable because we
    replace values with _CombinedConstVal. So the results should
    only be used for debugging/printing.
    """
    fg = forward_graph.ForwardGraph(allow_var_nodes=True)
    fg.add_nodes(outputs)
    new_outputs: list[graph.Node] = []
    next_items = list(fg.roots)
    replacements: dict[graph.Node, graph.OutputNode] = {}
    while next_items:
        forward_node = next_items.pop()
        if not forward_node.input_to:
            # This is a leaf node, so we can't combine it
            new_node = graph.OutputNode(
                forward_node.node.type,
                forward_node.node.from_op.name,
                {
                    k: replacements.get(v, v)
                    for k, v in forward_node.node.from_op.inputs.items()
                },
            )
            new_outputs.append(new_node)
            continue
        input_to = forward_node.input_to
        combinable_forward_nodes: set[forward_graph.ForwardNode] = set()

        # Find input forward nodes that are mergeable
        for fn in input_to:
            inputs = list(fn.node.from_op.inputs.values())
            if all(
                replacements.get(i, i) is forward_node.node
                or (
                    isinstance(i, graph.ConstNode)
                    and not isinstance(i.val, graph.OutputNode)
                )
                for i in inputs
            ):
                combinable_forward_nodes.add(fn)
            else:
                if fn not in next_items:
                    next_items.append(fn)

        # Group by op and merge
        name_key = (
            lambda n: n.node.from_op.name
            + "-"
            + "".join(
                "1" if isinstance(i, graph.ConstNode) else "0"
                for i in n.node.from_op.inputs.values()
            )
        )
        op_groups = itertools.groupby(
            sorted(combinable_forward_nodes, key=name_key), key=name_key
        )
        for k, group_nodes_iter in op_groups:
            group_nodes = list(group_nodes_iter)
            node0 = group_nodes[0]
            node0_inputs = list(node0.node.from_op.inputs.items())
            node_inputs: dict[str, forward_graph.ExecutableNode]
            if len(group_nodes) == 1:
                node_inputs = node0.node.from_op.inputs
            else:
                node_inputs = {}
                for input_name, input_node in node0_inputs:
                    if replacements.get(input_node, input_node) is forward_node.node:
                        node_inputs[input_name] = replacements.get(
                            input_node, input_node
                        )
                    else:
                        node_inputs[input_name] = graph.ConstNode(
                            input_node.type,
                            _CombinedConstVal(
                                [
                                    op.node.from_op.inputs[input_name].val
                                    for op in group_nodes
                                ]
                            ),
                        )
            new_inner_input_to: dict[
                forward_graph.ForwardNode, typing.Literal[True]
            ] = {}
            for o in group_nodes:
                new_inner_input_to.update(o.input_to)
            new_node = graph.OutputNode(
                node0.node.type, node0.node.from_op.name, node_inputs
            )
            for o in group_nodes:
                replacements[o.node] = new_node
            new_forward_node = forward_graph.ForwardNode(new_node)
            new_forward_node.input_to = new_inner_input_to
            if new_forward_node not in next_items:
                next_items.append(new_forward_node)

    return new_outputs


def node_expr_str_full(node: graph.Node) -> str:
    """Prints Node as an expression string, with full op names.

    This function is a copy/modification of of node_expr_str.
    """

    from weave.legacy.weave import partial_object

    if isinstance(node, graph.OutputNode):
        if node.from_op.name == "dict":
            return "dict(%s)" % ", ".join(
                "%s= %s" % (k, node_expr_str_full(v))
                for k, v in node.from_op.inputs.items()
            )
        elif node.from_op.name == "ArrowWeaveList-vectorizedDict":
            return "ArrowWeaveList-vectorizedDict(%s)" % ", ".join(
                "%s= %s" % (k, node_expr_str_full(v))
                for k, v in node.from_op.inputs.items()
            )
        if node.from_op.name == "gqlroot-wbgqlquery":
            query_hash = "_query_"  # TODO: make a hash from the query for idenity
            return f"{node.from_op.friendly_name}({query_hash})"
        elif node.from_op.name == "gqlroot-querytoobj":
            param_names = list(node.from_op.inputs.keys())
            const = node.from_op.inputs[param_names[2]]
            try:
                assert isinstance(const, graph.ConstNode)
                narrow_type = const.val
                assert isinstance(narrow_type, partial_object.PartialObjectType)
            except AssertionError:
                return (
                    f"{node_expr_str_full(node.from_op.inputs[param_names[0]])}."
                    f"querytoobj({node_expr_str_full(node.from_op.inputs[param_names[1]])}, ?)"
                )
            else:
                return (
                    f"{node_expr_str_full(node.from_op.inputs[param_names[0]])}."
                    f"querytoobj({node_expr_str_full(node.from_op.inputs[param_names[1]])},"
                    f" {narrow_type.keyless_weave_type_class()})"
                )

        param_names = list(node.from_op.inputs.keys())
        if all(
            [not isinstance(n, graph.OutputNode) for n in node.from_op.inputs.values()]
        ):
            return "%s(%s)" % (
                graph.opuri_full_name(node.from_op.name),
                ", ".join(
                    node_expr_str_full(node.from_op.inputs[n]) for n in param_names
                ),
                # node.type.simple_str()[:100],
            )
        if not param_names:
            return "%s()" % graph.opuri_full_name(node.from_op.name)
        # This puts in newlines, but not very well. Fix me :)
        return "%s\n  .%s(%s) " % (
            node_expr_str_full(node.from_op.inputs[param_names[0]]),
            graph.opuri_full_name(node.from_op.name),
            ", ".join(
                node_expr_str_full(node.from_op.inputs[n]) for n in param_names[1:]
            ),
            # ""
            # node.type.simple_str()[:100],
        )
    elif isinstance(node, graph.ConstNode):
        if isinstance(node.type, types.Function):
            res = node_expr_str_full(node.val)

            res = textwrap.indent(res, "  ")
            return res
        try:
            str_val = str(node.val)
            if len(str_val) > 50:
                return f"~{str_val[:50]}...~"
            return str_val
        except TypeError:
            # WARNING: This behavior means that sometimes this function
            # produces expressionions that JS can't parse (it happens when
            # we have Python Objects as values that have not yet been serialized)
            # TODO: fix
            return str(node.val)
    elif isinstance(node, graph.VarNode):
        return node.name
    elif isinstance(node, graph.VoidNode):
        return "<void>"
    else:
        return "**PARSE_ERROR**"


def to_assignment_form(
    outputs: list[graph.Node],
) -> dict[graph.OutputNode, graph.VarNode]:
    fg = forward_graph.ForwardGraph(allow_var_nodes=True)
    fg.add_nodes(outputs)
    new_nodes: dict[graph.Node, graph.OutputNode] = {}
    assignments: dict[graph.Node, graph.VarNode] = {}
    next_assignment_roots = list(fg.roots)
    var_id = 0
    while next_assignment_roots:
        forward_node = next_assignment_roots.pop()
        node = forward_node.node
        node_inputs: dict[str, graph.Node] = {}
        for k, v in node.from_op.inputs.items():
            if v in assignments:
                node_inputs[k] = assignments[v]
            else:
                node_inputs[k] = v
        new_node = graph.OutputNode(node.type, node.from_op.name, node_inputs)

        # walk linear graph section, swapping nodes for variables pointing
        # assignments we've already created
        while len(forward_node.input_to) == 1:
            next_forward_node = list(forward_node.input_to)[0]
            next_node = next_forward_node.node
            next_node_inputs: dict[str, graph.Node] = {}
            for k, v in next_node.from_op.inputs.items():
                if v is node:
                    next_node_inputs[k] = new_node
                elif v in assignments:
                    next_node_inputs[k] = assignments[v]
                else:
                    next_node_inputs[k] = v
            new_node = graph.OutputNode(
                next_node.type, next_node.from_op.name, next_node_inputs
            )
            forward_node = next_forward_node
            node = next_node

        # store in assignment
        new_nodes[node] = new_node
        assignments[node] = graph.VarNode(node.type, f"var{var_id}")
        var_id += 1

        # if we have more than one input_to, then create those as roots
        if len(forward_node.input_to) > 1:
            for input_to in forward_node.input_to:
                if input_to not in next_assignment_roots:
                    next_assignment_roots.append(input_to)

        # else we have zero, this is terminal
    return {new_nodes[k]: v for k, v in assignments.items()}


def assignments_string(assignments: dict[graph.OutputNode, graph.VarNode]) -> str:
    return "\n".join(
        f"{str(var_node)} = {node_expr_str_full(expression)}"
        for expression, var_node in assignments.items()
    )



import importlib
from typing import Callable

import weave
from weave.trace.patcher import MultiPatcher, SymbolPatcher


def dspy_wrapper(name: str) -> Callable[[Callable], Callable]:
    def wrapper(fn: Callable) -> Callable:
        op = weave.op()(fn)
        op.name = name  # type: ignore
        return op

    return wrapper


def dspy_get_patched_lm_functions(
    base_symbol: str, lm_class_name: str
) -> list[SymbolPatcher]:
    patchable_functional_attributes = [
        "basic_request",
        "request",
        "__call__",
    ]
    return [
        SymbolPatcher(
            get_base_symbol=lambda: importlib.import_module(base_symbol),
            attribute_name=f"{lm_class_name}.basic_request",
            make_new_value=dspy_wrapper(f"dspy.{lm_class_name}.basic_request"),
        ),
        SymbolPatcher(
            get_base_symbol=lambda: importlib.import_module(base_symbol),
            attribute_name=f"{lm_class_name}.request",
            make_new_value=dspy_wrapper(f"dspy.{lm_class_name}.request"),
        ),
        SymbolPatcher(
            get_base_symbol=lambda: importlib.import_module(base_symbol),
            attribute_name=f"{lm_class_name}.__call__",
            make_new_value=dspy_wrapper(f"dspy.{lm_class_name}"),
        ),
    ]


patched_functions = [
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="Predict.__call__",
        make_new_value=dspy_wrapper("dspy.Predict"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="Predict.forward",
        make_new_value=dspy_wrapper("dspy.Predict.forward"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="TypedPredictor.__call__",
        make_new_value=dspy_wrapper("dspy.TypedPredictor"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="TypedPredictor.forward",
        make_new_value=dspy_wrapper("dspy.TypedPredictor.forward"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="Module.__call__",
        make_new_value=dspy_wrapper("dspy.Module"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="TypedChainOfThought.__call__",
        make_new_value=dspy_wrapper("dspy.TypedChainOfThought"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="Retrieve.__call__",
        make_new_value=dspy_wrapper("dspy.Retrieve"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="Retrieve.forward",
        make_new_value=dspy_wrapper("dspy.Retrieve.forward"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.evaluate.evaluate"),
        attribute_name="Evaluate.__call__",
        make_new_value=dspy_wrapper("dspy.evaluate.Evaluate"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="BootstrapFewShot.compile",
        make_new_value=dspy_wrapper("dspy.teleprompt.BootstrapFewShot.compile"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="COPRO.compile",
        make_new_value=dspy_wrapper("dspy.teleprompt.COPRO.compile"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="Ensemble.compile",
        make_new_value=dspy_wrapper("dspy.teleprompt.Ensemble.compile"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="BootstrapFinetune.compile",
        make_new_value=dspy_wrapper("dspy.teleprompt.BootstrapFinetune.compile"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="KNNFewShot.compile",
        make_new_value=dspy_wrapper("dspy.teleprompt.KNNFewShot.compile"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="MIPRO.compile",
        make_new_value=dspy_wrapper("dspy.teleprompt.MIPRO.compile"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="BootstrapFewShotWithRandomSearch.compile",
        make_new_value=dspy_wrapper(
            "dspy.teleprompt.BootstrapFewShotWithRandomSearch.compile"
        ),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="SignatureOptimizer.compile",
        make_new_value=dspy_wrapper("dspy.teleprompt.SignatureOptimizer.compile"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="BayesianSignatureOptimizer.compile",
        make_new_value=dspy_wrapper(
            "dspy.teleprompt.BayesianSignatureOptimizer.compile"
        ),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module(
            "dspy.teleprompt.signature_opt_typed"
        ),
        attribute_name="optimize_signature",
        make_new_value=dspy_wrapper(
            "dspy.teleprompt.signature_opt_typed.optimize_signature"
        ),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="BootstrapFewShotWithOptuna.compile",
        make_new_value=dspy_wrapper(
            "dspy.teleprompt.BootstrapFewShotWithOptuna.compile"
        ),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy.teleprompt"),
        attribute_name="LabeledFewShot.compile",
        make_new_value=dspy_wrapper("dspy.teleprompt.LabeledFewShot.compile"),
    ),
]

# Patch LM classes
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="AzureOpenAI"
)
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="OpenAI"
)
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="Cohere"
)
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="Clarifai"
)
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="Google"
)
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="HFClientTGI"
)
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="HFClientVLLM"
)
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="Anyscale"
)
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="Together"
)
patched_functions += dspy_get_patched_lm_functions(
    base_symbol="dspy", lm_class_name="OllamaLocal"
)
patched_functions += [
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="Databricks.basic_request",
        make_new_value=dspy_wrapper("dspy.Databricks.basic_request"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="Databricks.__call__",
        make_new_value=dspy_wrapper("dspy.Databricks"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="ColBERTv2.__call__",
        make_new_value=dspy_wrapper("dspy.ColBERTv2"),
    ),
    SymbolPatcher(
        get_base_symbol=lambda: importlib.import_module("dspy"),
        attribute_name="Pyserini.__call__",
        make_new_value=dspy_wrapper("dspy.Pyserini"),
    ),
]

dspy_patcher = MultiPatcher(patched_functions)



import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave.language_features.tagging import tagged_value_type


def push_down_tags_from_container_type_to_element_type(
    container_type: types.Type,
) -> types.Type:
    if not isinstance(container_type, tagged_value_type.TaggedValueType):
        return container_type

    if not types.List().assign_type(container_type):
        raise ValueError(
            f"push_down_tags: expected container_type to be a list, got {container_type}"
        )

    container_tag_type = container_type.tag
    container_value_type = container_type.value.object_type  # type: ignore
    if isinstance(container_type.object_type, tagged_value_type.TaggedValueType):
        new_tag_type = types.TypedDict(
            {
                **container_tag_type.property_types,
                **container_type.object_type.tag.property_types,  # type: ignore
            }
        )
    else:
        new_tag_type = container_tag_type
    new_inner_type = tagged_value_type.TaggedValueType(
        new_tag_type, container_value_type
    )
    container_outer_type = container_type.value.__class__
    return container_outer_type(new_inner_type)  # type: ignore


def is_tagged_value_type(t: types.Type) -> bool:
    return (
        isinstance(t, tagged_value_type.TaggedValueType)
        or isinstance(t, types.Const)
        and isinstance(t.val_type, tagged_value_type.TaggedValueType)
    )


def unwrap_tags(
    t: types.Type,
) -> typing.Tuple[types.Type, typing.Callable[[types.Type], types.Type]]:
    # unwrap the tags from a type until we hit the first untagged type.
    # return the untagged type and return a function that rewraps it or
    # another type in the original tag tree.

    # see test_tagging.py::test_unwrap_rewrap_tags for examples.

    def wrap_outer(unwrapped: types.Type) -> types.Type:
        if isinstance(t, tagged_value_type.TaggedValueType):
            return tagged_value_type.TaggedValueType(t.tag, unwrapped)
        return unwrapped

    if isinstance(t, tagged_value_type.TaggedValueType):
        unwrapped, wrap_inner = unwrap_tags(t.value)
    else:
        unwrapped, wrap_inner = t, lambda x: x

    return unwrapped, lambda unwrapped_inner: wrap_outer(wrap_inner(unwrapped_inner))



from weave.legacy.weave import context_state as _context_state

_loading_builtins_token = _context_state.set_loading_built_ins()

from weave.legacy.weave.arrow.arrow import *
from weave.legacy.weave.arrow.convert import *
from weave.legacy.weave.arrow.list_ import *
from weave.legacy.weave.ops_arrow import ops
from weave.legacy.weave.ops_arrow.list_range import range
from weave.legacy.weave.ops_arrow.vectorize import vectorize

_context_state.clear_loading_built_ins(_loading_builtins_token)



import pytest

from weave.trace_server import trace_server_interface as tsi
from weave.trace_server.orm import (
    Column,
    ParamBuilder,
    Table,
    _transform_external_field_to_internal_field,
    combine_conditions,
)


def test_parambuilder_clickhouse():
    pb = ParamBuilder(database_type="clickhouse")
    name = pb.add("bar", "foo", "String")
    assert name == "{foo:String}"
    name = pb.add(12, "bim")
    assert name == "{bim:UInt64}"
    assert pb.get_params() == {
        "foo": "bar",
        "bim": 12,
    }


def test_parambuilder_sqlite():
    pb = ParamBuilder(database_type="sqlite")
    placeholder = pb.add("bar", "foo", "String")
    assert placeholder == ":foo"
    placeholder = pb.add(12, "bim")
    assert placeholder == ":bim"
    assert pb.get_params() == {
        "foo": "bar",
        "bim": 12,
    }


def test_combine_conditions():
    with pytest.raises(ValueError):
        combine_conditions([], "NOT")

    assert combine_conditions([], "AND") == ""
    assert combine_conditions(["foo = 'bar'"], "AND") == "foo = 'bar'"
    assert combine_conditions(["foo = 'bar'"], "OR") == "foo = 'bar'"
    assert (
        combine_conditions(["foo = 'bar'", "bim = 12"], "AND")
        == "((foo = 'bar') AND (bim = 12))"
    )
    assert (
        combine_conditions(["foo = 'bar'", "bim = 12"], "OR")
        == "((foo = 'bar') OR (bim = 12))"
    )


def test_transform_external_field_to_internal_field():
    all_columns = ["id", "creator", "payload_dump"]
    json_columns = ["payload"]

    # Transforming a column that doesn't exist should raise
    with pytest.raises(ValueError):
        _transform_external_field_to_internal_field("foo", all_columns, json_columns)

    result = _transform_external_field_to_internal_field(
        "id", all_columns, json_columns
    )
    assert result[0] == "id"
    assert result[2] == {"id"}
    result = _transform_external_field_to_internal_field(
        "payload", all_columns, json_columns
    )
    assert result[0] == "payload_dump"
    assert result[2] == {"payload_dump"}
    pb = ParamBuilder(prefix="pb", database_type="sqlite")
    result = _transform_external_field_to_internal_field(
        "payload.address", all_columns, json_columns, param_builder=pb
    )
    assert result[0] == "json_extract(payload_dump, :pb_0)"
    assert result[2] == {"payload_dump"}
    pb = ParamBuilder(prefix="pb", database_type="clickhouse")
    result = _transform_external_field_to_internal_field(
        "payload.address", all_columns, json_columns, param_builder=pb
    )
    assert result[0] == "toString(JSON_VALUE(payload_dump, {pb_0:String}))"
    assert result[2] == {"payload_dump"}


def test_table_create_sql():
    table = Table(
        "users",
        [
            Column("id", "string"),
            Column("creator", "string", nullable=True),
            Column("payload", "json", db_name="payload_dump"),
        ],
    )
    assert (
        table.create_sql()
        == "CREATE TABLE IF NOT EXISTS users (\n    id TEXT,\n    creator TEXT,\n    payload_dump TEXT\n)"
    )


def test_table_drop_sql():
    table = Table(
        "users",
        [
            Column("id", "string"),
            Column("creator", "string", nullable=True),
            Column("payload", "json", db_name="payload_dump"),
        ],
    )
    assert table.drop_sql() == "DROP TABLE IF EXISTS users"


def test_select_basic():
    table = Table(
        "users",
        [
            Column("id", "string"),
            Column("creator", "string", nullable=True),
            Column("payload", "json", db_name="payload_dump"),
        ],
    )
    select = table.select()
    select = select.limit(10)
    prepared = select.prepare(database_type="clickhouse")
    assert (
        prepared.sql
        == """SELECT id, creator, payload_dump
FROM users
LIMIT {limit:UInt64}"""
    )
    assert prepared.parameters == {"limit": 10}
    assert prepared.fields == ["id", "creator", "payload_dump"]

    prepared = select.prepare(database_type="sqlite")
    assert (
        prepared.sql
        == """SELECT id, creator, payload_dump
FROM users
LIMIT :limit"""
    )
    assert prepared.parameters == {"limit": 10}
    assert prepared.fields == ["id", "creator", "payload_dump"]


def test_select_fields():
    table = Table(
        "users",
        [
            Column("id", "string"),
            Column("creator", "string", nullable=True),
            Column("payload", "json", db_name="payload_dump"),
        ],
    )
    select = table.select()
    select = select.fields(["id", "payload.address"])

    # More complicated than normal usage to fix the ParamBuilder prefix.
    pb = ParamBuilder(prefix="pb", database_type="sqlite")
    prepared = select.prepare(database_type="sqlite", param_builder=pb)
    assert (
        prepared.sql
        == """SELECT id, json_extract(payload_dump, :pb_0)
FROM users"""
    )
    assert prepared.parameters == {"pb_0": '$."address"'}
    assert prepared.fields == ["id", "payload.address"]

    pb = ParamBuilder(prefix="pb", database_type="clickhouse")
    prepared = select.prepare(database_type="clickhouse", param_builder=pb)
    assert (
        prepared.sql
        == """SELECT id, toString(JSON_VALUE(payload_dump, {pb_0:String}))
FROM users"""
    )
    assert prepared.parameters == {"pb_0": '$."address"'}
    assert prepared.fields == ["id", "payload.address"]


def test_join():
    table1 = Table(
        "users",
        [
            Column("id", "string"),
            Column("creator", "string", nullable=True),
            Column("payload", "json", db_name="payload_dump"),
        ],
    )
    table2 = Table(
        "roles",
        [
            Column("id", "string"),
            Column("name", "string"),
        ],
    )

    select = (
        table1.select()
        .fields(["id", "name"])
        .join(
            table2,
            tsi.Query(
                **{
                    "$expr": {
                        "$eq": [
                            {"$getField": "table1.id"},
                            {"$getField": "table2.id"},
                        ]
                    }
                }
            ),
        )
    )

    prepared = select.prepare(database_type="clickhouse")

    assert (
        prepared.sql
        == """SELECT id, name
FROM users
JOIN roles ON (table1.id = table2.id)"""
    )


def test_join_with_join_type():
    table1 = Table(
        "users",
        [
            Column("id", "string"),
            Column("creator", "string", nullable=True),
            Column("payload", "json", db_name="payload_dump"),
        ],
    )
    table2 = Table(
        "roles",
        [
            Column("id", "string"),
            Column("name", "string"),
        ],
    )

    select = (
        table1.select()
        .fields(["id", "name"])
        .join(
            table2,
            tsi.Query(
                **{
                    "$expr": {
                        "$eq": [
                            {"$getField": "table1.id"},
                            {"$getField": "table2.id"},
                        ]
                    }
                }
            ),
            "inner",
        )
    )

    prepared = select.prepare(database_type="clickhouse")

    assert (
        prepared.sql
        == """SELECT id, name
FROM users
inner JOIN roles ON (table1.id = table2.id)"""
    )


def test_group_by():
    table = Table(
        "users",
        [
            Column("id", "string"),
            Column("creator", "string", nullable=True),
            Column("payload", "json", db_name="payload_dump"),
        ],
    )

    select = table.select().fields(["id", "creator"]).group_by(["id", "creator"])

    prepared = select.prepare(database_type="clickhouse")

    assert (
        prepared.sql
        == """SELECT id, creator
FROM users
GROUP BY id, creator"""
    )
